import logging

from cmakelang import lex
from cmakelang.parse.common import KwargBreaker
from cmakelang.parse.simple_nodes import CommentNode
from cmakelang.parse.argument_nodes import (
    KeywordGroupNode, PositionalGroupNode, PositionalParser, StandardArgTree)
from cmakelang.parse.additional_nodes import PatternNode
from cmakelang.parse.util import (
    WHITESPACE_TOKENS,
    get_first_semantic_token,
    get_normalized_kwarg,
    should_break)

logger = logging.getLogger(__name__)


def parse_install_targets_sub(ctx, tokens, breakstack):
  """
    Parse the inner kwargs of an ``install(TARGETS)`` command. This is common
    logic for ARCHIVE, LIBRARY, RUNTIME, etc.
  :see: https://cmake.org/cmake/help/v3.14/command/install.html#targets
  """
  return StandardArgTree.parse(
      ctx, tokens,
      npargs='*',
      kwargs={
          "DESTINATION": PositionalParser(1),
          "PERMISSIONS": PositionalParser('+'),
          "CONFIGURATIONS": PositionalParser('+'),
          "COMPONENT": PositionalParser(1),
          "NAMELINK_COMPONENT": PositionalParser(1),
      },
      flags=[
          "OPTIONAL",
          "EXCLUDE_FROM_ALL",
          "NAMELINK_ONLY",
          "NAMELINK_SKIP"
      ],
      breakstack=breakstack)


def parse_install_targets(ctx, tokens, breakstack):
  """
  ::

    install(TARGETS targets... [EXPORT <export-name>]
            [[ARCHIVE|LIBRARY|RUNTIME|OBJECTS|FRAMEWORK|BUNDLE|
              PRIVATE_HEADER|PUBLIC_HEADER|RESOURCE]
             [DESTINATION <dir>]
             [PERMISSIONS permissions...]
             [CONFIGURATIONS [Debug|Release|...]]
             [COMPONENT <component>]
             [NAMELINK_COMPONENT <component>]
             [OPTIONAL] [EXCLUDE_FROM_ALL]
             [NAMELINK_ONLY|NAMELINK_SKIP]
            ] [...]
            [INCLUDES DESTINATION [<dir> ...]]
            )

  :see: https://cmake.org/cmake/help/v3.14/command/install.html#targets
  """
  kwargs = {
      "TARGETS": PositionalParser('+'),
      "EXPORT": PositionalParser(1),
      "INCLUDES": PositionalParser('+', flags=["DESTINATION"]),
      # Common kwargs
      "DESTINATION": PositionalParser(1),
      "PERMISSIONS": PositionalParser('+'),
      "CONFIGURATIONS": PositionalParser('+'),
      "COMPONENT": PositionalParser(1),
      "NAMELINK_COMPONENT": PositionalParser(1),
  }
  flags = (
      "OPTIONAL",
      "EXCLUDE_FROM_ALL",
      "NAMELINK_ONLY",
      "NAMELINK_SKIP"
  )
  designated_kwargs = (
      "ARCHIVE", "LIBRARY", "RUNTIME", "OBJECTS", "FRAMEWORK",
      "BUNDLE", "PRIVATE_HEADER", "PUBLIC_HEADER", "RESOURCE"
  )

  # NOTE(josh): from here on, code is essentially StandardArgTree.parse(),
  # except that
  # we cannot break on the common subset of kwargs in the breakstack because
  # they are valid kwargs for the subtrees (ARCHIVE, LIBRARY, etc) as well as
  # the primary tree
  tree = StandardArgTree()

  # If it is a whitespace token then put it directly in the parse tree at
  # the current depth
  while tokens and tokens[0].type in WHITESPACE_TOKENS:
    tree.children.append(tokens.pop(0))
    continue

  # ARCHIVE, LIBRARY, RUNTIME, subtrees etc only break on the start of
  # another subtree, or on "INCLUDES DESTINATION"
  subtree_breakstack = breakstack + [KwargBreaker(
      list(designated_kwargs) + ["INCLUDES"]
  )]

  # kwargs at this tree depth break on other kwargs or flags
  kwarg_breakstack = breakstack + [KwargBreaker(
      list(kwargs.keys()) + list(designated_kwargs) + list(flags)
  )]

  # and flags at this depth break only on kwargs
  positional_breakstack = breakstack + [KwargBreaker(
      list(kwargs.keys()) + list(designated_kwargs)
  )]

  while tokens:
    # Break if the next token belongs to a parent parser, i.e. if it
    # matches a keyword argument of something higher in the stack, or if
    # it closes a parent group.
    if should_break(tokens[0], breakstack):
      break

    # If it is a whitespace token then put it directly in the parse tree at
    # the current depth
    if tokens[0].type in WHITESPACE_TOKENS:
      tree.children.append(tokens.pop(0))
      continue

    # If it's a comment, then add it at the current depth
    if tokens[0].type in (lex.TokenType.COMMENT,
                          lex.TokenType.BRACKET_COMMENT):
      child = CommentNode()
      tree.children.append(child)
      child.children.append(tokens.pop(0))
      continue

    ntokens = len(tokens)
    # NOTE(josh): each flag is also stored in kwargs as with a positional parser
    # of size zero. This is a legacy thing that should be removed, but for now
    # just make sure we check flags first.
    word = get_normalized_kwarg(tokens[0])
    if word in designated_kwargs:
      subtree = KeywordGroupNode.parse(
          ctx, tokens, word, parse_install_targets, subtree_breakstack)
    elif word in kwargs:
      subtree = KeywordGroupNode.parse(
          ctx, tokens, word, kwargs[word], kwarg_breakstack)
    else:
      subtree = PositionalGroupNode.parse(
          ctx, tokens, '+', flags, positional_breakstack)

    assert len(tokens) < ntokens
    tree.children.append(subtree)
  return tree


def parse_install_files(ctx, tokens, breakstack):
  """
  ::
    install(<FILES|PROGRAMS> files...
            TYPE <type> | DESTINATION <dir>
            [PERMISSIONS permissions...]
            [CONFIGURATIONS [Debug|Release|...]]
            [COMPONENT <component>]
            [RENAME <name>] [OPTIONAL] [EXCLUDE_FROM_ALL])

  :see: https://cmake.org/cmake/help/v3.14/command/install.html#files
  """
  return StandardArgTree.parse(
      ctx, tokens,
      npargs='*',
      kwargs={
          "FILES": PositionalParser('+'),
          "PROGRAMS": PositionalParser('+'),
          "TYPE": PositionalParser(1),
          "DESTINATION": PositionalParser(1),
          "PERMISSIONS": PositionalParser('+'),
          "CONFIGURATIONS": PositionalParser('+'),
          "COMPONENT": PositionalParser(1),
          "RENAME": PositionalParser(1),
      },
      flags=[
          "OPTIONAL",
          "EXCLUDE_FROM_ALL",
      ],
      breakstack=breakstack)


def parse_install_directory(ctx, tokens, breakstack):
  """
  ::

    install(DIRECTORY dirs...
            TYPE <type> | DESTINATION <dir>
            [FILE_PERMISSIONS permissions...]
            [DIRECTORY_PERMISSIONS permissions...]
            [USE_SOURCE_PERMISSIONS] [OPTIONAL] [MESSAGE_NEVER]
            [CONFIGURATIONS [Debug|Release|...]]
            [COMPONENT <component>] [EXCLUDE_FROM_ALL]
            [FILES_MATCHING]
            [[PATTERN <pattern> | REGEX <regex>]
             [EXCLUDE] [PERMISSIONS permissions...]] [...])

  :see: https://cmake.org/cmake/help/v3.14/command/install.html#directory
  """
  return StandardArgTree.parse(
      ctx, tokens,
      npargs='*',
      kwargs={
          "DIRECTORY": PositionalParser('+'),
          "TYPE": PositionalParser(1),
          "DESTINATION": PositionalParser(1),
          "FILE_PERMISSIONS": PositionalParser('+'),
          "DIRECTORY_PERMISSIONS": PositionalParser('+'),
          "CONFIGURATIONS": PositionalParser('+'),
          "COMPONENT": PositionalParser(1),
          "RENAME": PositionalParser(1),
          "PATTERN": PatternNode.parse,
          "REGEX": PatternNode.parse,
      },
      flags=[
          "USER_SOURCE_PERMISSIONS",
          "OPTIONAL",
          "MESSAGE_NEVER",
          "FILES_MATCHING",
      ],
      breakstack=breakstack)


def parse_install_script(ctx, tokens, breakstack):
  """
  ::

    install([[SCRIPT <file>] [CODE <code>]]
            [COMPONENT <component>] [EXCLUDE_FROM_ALL] [...])

  :see: https://cmake.org/cmake/help/v3.14/command/install.html#custom-installation-logic
  """
  return StandardArgTree.parse(
      ctx, tokens,
      npargs='*',
      kwargs={
          "SCRIPT": PositionalParser(1),
          "CODE": PositionalParser(1),
          "COMPONENT": PositionalParser(1),
      },
      flags=[
          "EXCLUDE_FROM_ALL"
      ],
      breakstack=breakstack)


def parse_install_export(ctx, tokens, breakstack):
  """
  ::

    install(EXPORT <export-name> DESTINATION <dir>
            [NAMESPACE <namespace>] [[FILE <name>.cmake]|
            [PERMISSIONS permissions...]
            [CONFIGURATIONS [Debug|Release|...]]
            [EXPORT_LINK_INTERFACE_LIBRARIES]
            [COMPONENT <component>]
            [EXCLUDE_FROM_ALL])

  :see: https://cmake.org/cmake/help/v3.14/command/install.html#installing-exports
  """
  return StandardArgTree.parse(
      ctx, tokens,
      npargs='*',
      kwargs={
          "EXPORT": PositionalParser(1),
          "DESTINATION": PositionalParser(1),
          "NAMESPACE": PositionalParser(1),
          "FILE": PositionalParser(1),
          "PERMISSIONS": PositionalParser('+'),
          "CONFIGURATIONS": PositionalParser('+'),
          "COMPONENT": PositionalParser(1),
      },
      flags=[
          "EXCLUDE_FROM_ALL"
      ],
      breakstack=breakstack)


def parse_install(ctx, tokens, breakstack):
  """
  The ``install()`` command has multiple different forms, implemented
  by different functions. The forms are indicated by the first token
  and are:

  * TARGETS
  * FILES
  * DIRECTORY
  * SCRIPT
  * CODE
  * EXPORT

  :see: https://cmake.org/cmake/help/v3.0/command/install.html
  """

  descriminator_token = get_first_semantic_token(tokens)
  if descriminator_token is None:
    logger.warning("Invalid install() command at %s", tokens[0].get_location())
    return StandardArgTree.parse(ctx, tokens, npargs='*', kwargs={}, flags=[],
                                 breakstack=breakstack)

  descriminator = descriminator_token.spelling.upper()
  parsemap = {
      "TARGETS": parse_install_targets,
      "FILES": parse_install_files,
      "PROGRAMS": parse_install_files,
      "DIRECTORY": parse_install_directory,
      "SCRIPT": parse_install_script,
      "CODE": parse_install_script,
      "EXPORT": parse_install_export
  }
  if descriminator not in parsemap:
    logger.warning("Invalid install form \"%s\" at %s", descriminator,
                   tokens[0].location())
    return StandardArgTree.parse(ctx, tokens, npargs='*', kwargs={}, flags=[],
                                 breakstack=breakstack)

  return parsemap[descriminator](ctx, tokens, breakstack)


def populate_db(parse_db):
  parse_db["install"] = parse_install



# -*- coding: utf-8 -*-
# pylint: disable=too-many-lines
from __future__ import print_function
from __future__ import unicode_literals

import contextlib

from cmakelang import common


class MockEverything(object):
  """Dummy object which implements any interface by mocking all functions
     with an empty implementation that returns None"""

  # pylint: disable=unused-argument
  def _dummy(self, *_args, **_kwargs):
    return

  def __getattr__(self, _name):
    return self._dummy


class ParseContext(object):
  """Global context passed through every function in the parse stack."""

  def __init__(self, parse_db=None, lint_ctx=None, config=None):
    if parse_db is None:
      from cmakelang.parse.funs import get_parse_db
      parse_db = get_parse_db()
    self.parse_db = parse_db

    if lint_ctx is None:
      lint_ctx = MockEverything()
    self.lint_ctx = lint_ctx

    if config is None:
      from cmakelang import configuration
      config = configuration.Configuration()
    self.config = config

    # List of currently open parse nodes. Only used by nodes below
    # the statement level.
    self.argstack = []

  @contextlib.contextmanager
  def pusharg(self, node):
    self.argstack.append(node)
    yield None
    if not self.argstack:
      raise common.InternalError(
          "Unexpected empty argstack, expected {}".format(node))

    if self.argstack[-1] is not node:
      raise common.InternalError(
          "Unexpected node {} on argstack, expecting {}"
          .format(self.argstack[-1], node))

    self.argstack.pop(-1)


def parse(tokens, ctx=None):
  """
  digest tokens, then layout the digested blocks.
  """
  if ctx is None:
    ctx = ParseContext()
  from cmakelang.parse.body_nodes import BodyNode
  return BodyNode.consume(ctx, tokens)



"""
Command specifications for cmake built-in commands.
"""

from __future__ import unicode_literals

from cmakelang.parse.funs import standard_builtins, standard_modules
from cmakelang.parse.util import CommandSpec

CANONICAL_SPELLINGS = {
    "FetchContent_Declare",
    "FetchContent_MakeAvailable",
    "FetchContent_GetProperties",
    "FetchContent_Populate",
    "ExternalProject_Add",
    "ExternalProject_Add_Step",
    "ExternalProject_Add_StepDependencies",
    "ExternalProject_Add_StepTargets",
    "ExternalProject_Get_Property",
}


def get_default_config():
  """
  Return the default per-command configuration database
  """
  per_command = {}
  for spelling in CANONICAL_SPELLINGS:
    per_command[spelling.lower()] = {
        "spelling": spelling
    }

  return per_command


def get_fn_spec():
  """
  Return a dictionary mapping cmake function names to a dictionary containing
  kwarg specifications.
  """

  fn_spec = CommandSpec('<root>')
  for grouping in (standard_builtins, standard_modules):
    for funname, spec in sorted(grouping.FUNSPECS.items()):
      fn_spec.add(funname, **spec)
  return fn_spec


# pylint: disable=bad-continuation
# pylint: disable=too-many-lines



# -*- coding: utf-8 -*-
"""
Analyze the naming conventions by recording all variable assignments and
classifying them according to context.
"""

from __future__ import unicode_literals

import argparse
import io
import logging
import os
import re
import sys

import cmakelang
from cmakelang import common
from cmakelang.format import __main__
from cmakelang import configuration
from cmakelang import lex
from cmakelang import parse

from cmakelang.lex import TokenType
from cmakelang.lint.basic_checker import (
    Scope, get_scope_of_assignment, find_statements_in_subtree)

logger = logging.getLogger(__name__)


class NameCollector(object):
  def __init__(self):
    self.varnames = []

  def collect_names(self, parse_tree):
    for stmt in find_statements_in_subtree(parse_tree, ["set", "list"]):
      scope = get_scope_of_assignment(stmt)
      if stmt.get_funname() == "set":
        token = stmt.argtree.varname
      else:
        token = stmt.argtree.parg_groups[0].get_tokens(kind="semantic")[1]
      self.varnames.append((scope, token.spelling))

    for stmt in find_statements_in_subtree(parse_tree, ["foreach"]):
      tokens = stmt.get_semantic_tokens()
      tokens.pop(0)  # statement name "foreach"
      tokens.pop(0)  # lparen
      token = tokens.pop(0)  # loopvar
      self.varnames.append((Scope.LOOP, token.spelling))

    for stmt in find_statements_in_subtree(parse_tree, ["function", "macro"]):
      tokens = stmt.argtree.get_semantic_tokens()[1:]
      for token in tokens:  # named arguments
        if token.type is TokenType.RIGHT_PAREN:
          break
        if token.type is not TokenType.WORD:
          logger.warning(
              "Unexpected token %s at %s", token.spelling, token.get_location())
        else:
          self.varnames.append((Scope.ARGUMENT, token.spelling))


def setup_argparse(argparser):
  argparser.add_argument('-v', '--version', action='version',
                         version=cmakelang.__version__)
  argparser.add_argument(
      '-l', '--log-level', default="info",
      choices=["error", "warning", "info", "debug"])

  argparser.add_argument(
      '-o', '--outfile-path', default=None,
      help='Write errors to this file. Default is stdout.')

  argparser.add_argument('infilepaths', nargs='*')


USAGE_STRING = """
analyze_naming [-h] [-o OUTFILE_PATH] infilepath [infilepath ...]
"""


def inner_main():
  """Parse arguments, open files, start work."""
  logging.basicConfig(level=logging.INFO, format="%(levelname)s %(message)s")

  argparser = argparse.ArgumentParser(
      description=__doc__,
      formatter_class=argparse.RawDescriptionHelpFormatter,
      usage=USAGE_STRING)

  setup_argparse(argparser)
  args = argparser.parse_args()
  logging.getLogger().setLevel(getattr(logging, args.log_level.upper()))

  if args.outfile_path is None:
    args.outfile_path = '-'

  if '-' in args.infilepaths:
    assert len(args.infilepaths) == 1, \
        "You cannot mix stdin as an input with other input files"

  if args.outfile_path == '-':
    outfile = io.open(os.dup(sys.stdout.fileno()),
                      mode='w', encoding="utf-8", newline='')
  else:
    outfile = io.open(args.outfile_path, 'w', encoding="utf-8", newline='')

  returncode = 0

  cfg = configuration.Configuration()
  collector = NameCollector()
  for infile_path in args.infilepaths:
    # NOTE(josh): have to load config once for every file, because we may pick
    # up a new config file location for each path
    if infile_path == '-':
      infile_path = os.dup(sys.stdin.fileno())

    try:
      infile = io.open(
          infile_path, mode='r', encoding=cfg.encode.input_encoding, newline='')
    except (IOError, OSError):
      logger.error("Failed to open %s for read", infile_path)
      returncode = 1
      continue

    try:
      with infile:
        infile_content = infile.read()
    except UnicodeDecodeError:
      logger.error(
          "Unable to read %s as %s", infile_path, cfg.encode.input_encoding)
      returncode = 1
      continue

    tokens = lex.tokenize(infile_content)
    parse_db = parse.funs.get_parse_db()
    ctx = parse.ParseContext(parse_db, config=cfg)
    parse_tree = parse.parse(tokens, ctx)
    parse_tree.build_ancestry()
    collector.collect_names(parse_tree)

  regexes = [re.compile(pattern) for pattern in [
      r"[A-Z][A-Z0-9_]+",  # upper snake-case
      r"[a-z][a-z0-9_]+",  # lower snake-case
      r"_[A-Z0-9_]+",  # upper snake-case with underscore prefix
      r"_[a-z0-9_]+",  # lower snake-case with underscore prefix
  ]]

  outmap = {}
  patmap = {}
  for scope, varname in sorted(collector.varnames):
    if scope not in outmap:
      outmap[scope] = {}

    if scope not in patmap:
      patmap[scope] = {}
      for regex in regexes:
        patmap[scope][str(regex)] = 0
      patmap[scope]["other"] = 0

    for regex in regexes:
      if regex.match(varname):
        patmap[scope][str(regex)] += 1
        break
    else:
      patmap[scope]["other"] += 1

    if varname not in outmap[scope]:
      outmap[scope][varname] = 0
    outmap[scope][varname] += 1

  for scope, countmap in sorted(outmap.items()):
    outfile.write("\n{}\n{}\n".format(scope.name, "=" * len(scope.name)))
    for varname, count in sorted(countmap.items()):
      outfile.write("{}: {}\n".format(varname, count))

  for scope, countmap in sorted(patmap.items()):
    outfile.write("\n{}\n{}\n".format(scope.name, "=" * len(scope.name)))
    for varname, count in sorted(countmap.items()):
      outfile.write("{}: {}\n".format(varname, count))

  outfile.close()
  return returncode


def main():
  try:
    return inner_main()
  except SystemExit:
    return 0
  except common.UserError as ex:
    logger.fatal(ex.msg)
    return 1
  except common.InternalError as ex:
    logger.exception(ex.msg)
    return 2
  except AssertionError as ex:
    logger.exception(
        "An internal error occured. Please consider filing a bug report at "
        "github.com/cheshirekow/cmakelang/issues")
    return 2
  except:  # pylint: disable=bare-except
    logger.exception(
        "An internal error occured. Please consider filing a bug report at "
        "github.com/cheshirekow/cmakelang/issues")
    return 2


if __name__ == "__main__":
  sys.exit(main())



"""
Generate configuration details docpage
"""

from __future__ import unicode_literals

import argparse
import io
import os
import pprint
import re
import sys
import tempfile

from cmakelang import configuration
from cmakelang import config_util


def parse_sourcefile(infile):
  ruler = re.compile(r"[=\-\^~]+")
  lines = infile.read().split("\n")

  out = {}
  stack = [out]
  sepmap = {
      "=": 0,
      "-": 1,
      "^": 2,
      "~": 3
  }

  varname = None
  buf = []

  lineiter = enumerate(lines)
  for idx, line in lineiter:
    if ruler.match(line):
      if lines[idx + 2] == line:
        # Section heading
        if varname:
          stack[-1][varname] = "\n".join(buf)
          varname = None
        buf = []

        title = lines[idx + 1]
        newdict = {}
        parent_idx = sepmap[line[0]]
        stack[parent_idx][title] = newdict
        stack = stack[:parent_idx + 1]
        stack.append(newdict)
        next(lineiter)
        next(lineiter)
    elif line and ruler.match(lines[idx + 1]):
      # Variable heading
      if varname:
        stack[-1][varname] = "\n".join(buf[:-1])
        varname = None
      buf = []
      varname = line
      next(lineiter)
    else:
      buf.append(line)

  if varname:
    stack[-1][varname] = "\n".join(buf[:-1])
    buf = []
    varname = None

  return out


SEPARATORS = ["=", "-", "^", "~"]


def get_command_line(descr):
  argparser = argparse.ArgumentParser(usage="")
  descr.add_to_argparse(argparser)
  sys.stdout.flush()
  stdout_copy = os.dup(sys.stdout.fileno())
  with tempfile.NamedTemporaryFile(delete=False) as outfile:
    tmppath = outfile.name
    os.dup2(outfile.fileno(), sys.stdout.fileno())
    try:
      argparser.parse_args(["--help"])
    except SystemExit:
      pass
  sys.stdout.flush()
  os.dup2(stdout_copy, sys.stdout.fileno())
  with io.open(tmppath, "r", encoding="utf-8") as infile:
    content = infile.read()
  os.unlink(tmppath)
  return "\n".join(content.split("\n")[4:])


ROOT_CONFIG = None


def get_config_example(root, obj, descr):
  default_value = descr.__get__(obj, None)
  descr.__set__(obj, default_value)
  buf = io.StringIO()
  root.dump(buf, with_defaults=False)
  descr.unset(obj)
  return buf.getvalue()


def write_outfile(outfile, data, obj, name=None, depth=0, root=None):
  if root is None:
    root = obj

  helptext = obj.__class__.__doc__.strip()
  title, helptext = (helptext + "\n").split("\n", 1)

  if not title and name:
    title = name

  outfile.write(SEPARATORS[depth] * len(title))
  outfile.write("\n")
  outfile.write(title)
  outfile.write("\n")
  outfile.write(SEPARATORS[depth] * len(title))
  outfile.write("\n\n")

  if helptext.strip():
    outfile.write("\n")
    outfile.write(helptext)

  ppr = pprint.PrettyPrinter(indent=2, width=72)
  for descr in obj._field_registry:  # pylint: disable=protected-access
    if isinstance(descr, config_util.SubtreeDescriptor):
      subobj = descr.__get__(obj, type(obj))
      write_outfile(
          outfile, data.get(descr.name, {}), subobj, descr.name, depth + 1,
          root)
    elif isinstance(descr, config_util.FieldDescriptor):
      outfile.write(".. _{}:\n\n".format(descr.name))
      outfile.write(descr.name)
      outfile.write("\n")
      outfile.write("=" * len(descr.name))
      outfile.write("\n\n")

      if descr.helptext:
        outfile.write(descr.helptext.strip())
        outfile.write("\n\n")

      heading = "default value:"
      outfile.write("{}\n{}\n\n".format(heading, "-" * len(heading)))
      outfile.write(".. code::\n\n")
      for line in ppr.pformat(descr.default_value).split("\n"):
        outfile.write("  {}\n".format(line))
      outfile.write("\n")

      if descr.name in data:
        heading = "detailed description:"
        outfile.write("{}\n{}\n\n".format(heading, "-" * len(heading)))
        outfile.write(data[descr.name].strip())
        outfile.write("\n\n")

      cmdline = get_command_line(descr)
      if cmdline:
        heading = "command-line option:"
        outfile.write("{}\n{}\n\n".format(heading, "-" * len(heading)))
        outfile.write(".. code:: \n\n")
        for line in cmdline.split("\n"):
          outfile.write("   {}\n".format(line))
        outfile.write("\n")

      config = get_config_example(root, obj, descr)
      if config:
        heading = "config-file entry:"
        outfile.write("{}\n{}\n\n".format(heading, "-" * len(heading)))
        outfile.write(".. code:: \n\n")
        for line in config.split("\n"):
          outfile.write("  {}\n".format(line))
        outfile.write("\n")

    else:
      raise RuntimeError("Field registry contains unknown descriptor")


def main():
  rootdir = os.sep.join(os.path.realpath(__file__).split(os.sep)[:-3])

  argparser = argparse.ArgumentParser(description=__doc__)
  argparser.add_argument(
      "--sourcefile",
      default=os.path.join(rootdir, "cmakelang/doc/configopts.rst"))
  argparser.add_argument(
      "outfile", nargs="?", default="-")
  args = argparser.parse_args()

  with io.open(args.sourcefile, "r", encoding="utf-8") as infile:
    data = parse_sourcefile(infile)

  if args.outfile == "-":
    args.outfile = os.dup(sys.stdout.fileno())

  with io.open(args.outfile, "w", encoding="utf-8") as outfile:
    outfile.write(".. _configopts:\n\n")
    write_outfile(
        outfile, data["global"], configuration.Configuration(), "global")


if __name__ == "__main__":
  main()



import logging

from cmakelang import lex
from cmakelang.parse.argument_nodes import (
    ArgGroupNode, PositionalGroupNode, StandardArgTree)
from cmakelang.parse.common import NodeType, TreeNode
from cmakelang.parse.simple_nodes import CommentNode

from cmakelang.parse.util import (
    WHITESPACE_TOKENS,
    get_tag,
    get_normalized_kwarg,
    iter_semantic_tokens,
    only_comments_and_whitespace_remain,
    PositionalSpec
)

logger = logging.getLogger(__name__)


def parse_add_executable_imported(ctx, tokens, breakstack):
  """
  ::

    add_executable(<name> IMPORTED [GLOBAL])

  :see: https://cmake.org/cmake/help/latest/command/add_executable.html
  """
  return StandardArgTree.parse(
      ctx, tokens, npargs='+',
      kwargs={},
      flags=["IMPORTED", "GLOBAL"],
      breakstack=breakstack)


def parse_add_executable_alias(ctx, tokens, breakstack):
  """
  ::
    add_executable(<name> ALIAS <target>)

    :see: https://cmake.org/cmake/help/latest/command/add_executable.html
  """
  return StandardArgTree.parse(
      ctx, tokens, npargs=3,
      kwargs={},
      flags=["ALIAS"],
      breakstack=breakstack)


def parse_add_executable_standard(ctx, tokens, breakstack, sortable):
  """
  ::

    add_executable(<name> [WIN32] [MACOSX_BUNDLE]
                   [EXCLUDE_FROM_ALL]
                   [source1] [source2 ...])

  :see: https://cmake.org/cmake/help/latest/command/add_executable.html#command:add_executable
  """
  # pylint: disable=too-many-statements

  parsing_name = 1
  parsing_flags = 2
  parsing_sources = 3

  tree = ArgGroupNode()

  # If it is a whitespace token then put it directly in the parse tree at
  # the current depth
  while tokens and tokens[0].type in WHITESPACE_TOKENS:
    tree.children.append(tokens.pop(0))
    continue

  state_ = parsing_name
  parg_group = None
  src_group = None
  active_depth = tree

  while tokens:
    # This parse function breaks on the first right paren, since parenthetical
    # groups are not allowed. A parenthesis might exist in a filename, but
    # if so that filename should be quoted so it wont show up as a RIGHT_PAREN
    # token.
    if tokens[0].type is lex.TokenType.RIGHT_PAREN:
      break

    # If it is a whitespace token then put it directly in the parse tree at
    # the current depth
    if tokens[0].type in WHITESPACE_TOKENS:
      active_depth.children.append(tokens.pop(0))
      continue

    # If it's a comment token not associated with an argument, then put it
    # directly into the parse tree at the current depth
    if tokens[0].type in (lex.TokenType.COMMENT,
                          lex.TokenType.BRACKET_COMMENT):
      if state_ > parsing_name:
        if get_tag(tokens[0]) in ("unsort", "unsortable"):
          sortable = False
        elif get_tag(tokens[0]) in ("unsort", "unsortable"):
          sortable = True
      child = TreeNode(NodeType.COMMENT)
      active_depth.children.append(child)
      child.children.append(tokens.pop(0))
      continue

    if state_ is parsing_name:
      token = tokens.pop(0)
      parg_group = PositionalGroupNode()
      parg_group.spec = PositionalSpec("+")
      active_depth = parg_group
      tree.children.append(parg_group)
      child = TreeNode(NodeType.ARGUMENT)
      child.children.append(token)
      CommentNode.consume_trailing(ctx, tokens, child)
      parg_group.children.append(child)
      state_ += 1
    elif state_ is parsing_flags:
      if get_normalized_kwarg(tokens[0]) in (
          "WIN32", "MACOSX_BUNDLE", "EXCLUDE_FROM_ALL"):
        token = tokens.pop(0)
        child = TreeNode(NodeType.FLAG)
        child.children.append(token)
        CommentNode.consume_trailing(ctx, tokens, child)
        parg_group.children.append(child)
      else:
        state_ += 1
        src_group = PositionalGroupNode(sortable=sortable, tags=["file-list"])
        src_group.spec = PositionalSpec("+")
        active_depth = src_group
        tree.children.append(src_group)
    elif state_ is parsing_sources:
      token = tokens.pop(0)
      child = TreeNode(NodeType.ARGUMENT)
      child.children.append(token)
      CommentNode.consume_trailing(ctx, tokens, child)
      src_group.children.append(child)

      if only_comments_and_whitespace_remain(tokens, breakstack):
        active_depth = tree

  return tree


def parse_add_executable(ctx, tokens, breakstack):
  """
  ``add_executable()`` has a couple of forms:

  * normal executables
  * imported executables
  * alias executables

  This function is just the dispatcher

  :see: https://cmake.org/cmake/help/latest/command/add_executable.html
  """

  semantic_iter = iter_semantic_tokens(tokens)
  # NOTE(josh): first token is always the name of the executable
  _ = next(semantic_iter, None)
  # Second token is usually the descriminator
  second_token = next(semantic_iter, None)

  if second_token is None:
    # All add_library() commands should have at least two arguments
    logger.warning("Invalid add_executable() command at %s",
                   tokens[0].get_location())
    return StandardArgTree.parse(ctx, tokens, npargs='*', kwargs={}, flags=[],
                                 breakstack=breakstack)

  descriminator = second_token.spelling.upper()
  parsemap = {
      "ALIAS": parse_add_executable_alias,
      "IMPORTED": parse_add_executable_imported
  }
  if descriminator in parsemap:
    return parsemap[descriminator](ctx, tokens, breakstack)

  # If the descriminator token might be a variable dereference, then it
  # might be hiding the descriminator... so we shouldn't infer
  # sortability unless it is a word that doesn't match any of the descriminator
  # flags
  sortable = True
  if "${" in second_token.spelling:
    sortable = False

  return parse_add_executable_standard(ctx, tokens, breakstack, sortable)


def populate_db(parse_db):
  parse_db["add_executable"] = parse_add_executable



# -*- coding: utf-8 -*-
# pylint: disable=too-many-lines
from __future__ import print_function
from __future__ import unicode_literals

import io
import sys

from cmakelang.parse.common import TreeNode


def dump_tree(nodes, outfile=None, indent=None):
  """
  Print a tree of node objects for debugging purposes
  """

  if indent is None:
    indent = ''

  if outfile is None:
    outfile = sys.stdout

  for idx, node in enumerate(nodes):
    outfile.write(indent)
    if idx + 1 == len(nodes):
      outfile.write('└─ ')
    else:
      outfile.write('├─ ')
    noderep = repr(node)
    if sys.version_info[0] < 3:
      # python2
      outfile.write(getattr(noderep, 'decode')('utf-8'))
    else:
      # python3
      outfile.write(noderep)
    outfile.write('\n')

    if not hasattr(node, 'children'):
      continue

    if idx + 1 == len(nodes):
      dump_tree(node.children, outfile, indent + '    ')
    else:
      dump_tree(node.children, outfile, indent + '│   ')


def dump_tree_tostr(nodes, indent=None):
  outfile = io.StringIO()
  dump_tree(nodes, outfile, indent)
  return outfile.getvalue()


def dump_tree_upto(nodes, history, outfile=None, indent=None):
  """
  Print a tree of node objects for debugging purposes
  """

  if indent is None:
    indent = ''

  if outfile is None:
    outfile = sys.stdout

  for idx, node in enumerate(nodes):
    outfile.write(indent)
    if idx + 1 == len(nodes):
      outfile.write('└─ ')
    else:
      outfile.write('├─ ')

    if history[0] is node:
      # ANSI red
      outfile.write("\u001b[31m")
      if sys.version_info[0] < 3:
        outfile.write(repr(node).decode('utf-8'))
      else:
        outfile.write(repr(node))
      # ANSI reset
      outfile.write("\u001b[0m")
    else:
      if sys.version_info[0] < 3:
        outfile.write(repr(node).decode('utf-8'))
      else:
        outfile.write(repr(node))
    outfile.write('\n')

    if not hasattr(node, 'children'):
      continue

    subhistory = history[1:]
    if not subhistory:
      continue

    if idx + 1 == len(nodes):
      dump_tree_upto(node.children, subhistory, outfile, indent + '    ')
    else:
      dump_tree_upto(node.children, subhistory, outfile, indent + '│   ')


def tree_string(nodes, history=None):
  outfile = io.StringIO()
  if history:
    dump_tree_upto(nodes, history, outfile)
  else:
    dump_tree(nodes, outfile)
  return outfile.getvalue()


def has_nontoken_children(node):
  if not hasattr(node, 'children'):
    return False

  for child in node.children:
    if isinstance(child, TreeNode):
      return True

  return False


def dump_tree_for_test(nodes, outfile=None, indent=None, increment=None):
  """
  Print a tree of node objects for debugging purposes
  """
  if increment is None:
    increment = '    '

  if indent is None:
    indent = ''

  if outfile is None:
    outfile = sys.stdout

  for node in nodes:
    if not isinstance(node, TreeNode):
      continue
    outfile.write(indent)
    outfile.write('({}, ['.format(node.node_type))
    if has_nontoken_children(node):
      outfile.write('\n')
      dump_tree_for_test(node.children, outfile, indent + increment, increment)
      outfile.write(indent)
    outfile.write(']),\n')


def test_string(nodes, indent=None, increment=None):
  if indent is None:
    indent = ' ' * 10
  if increment is None:
    increment = '    '

  outfile = io.StringIO()
  dump_tree_for_test(nodes, outfile, indent, increment)
  return outfile.getvalue()



from __future__ import unicode_literals
import contextlib
import difflib
import io
import os
import shutil
import subprocess
import sys
import unittest
import tempfile


class TestInvocations(unittest.TestCase):

  def __init__(self, *args, **kwargs):
    super(TestInvocations, self).__init__(*args, **kwargs)
    self.tempdir = None
    self.tempconfig = None

  def setUp(self):
    self.tempdir = tempfile.mkdtemp(prefix='cmakeformattest_')
    thisdir = os.path.realpath(os.path.dirname(__file__))
    parentdir = os.path.dirname(thisdir)

    self.env = {
        'PYTHONPATH': os.path.dirname(parentdir)
    }

    configpath = os.path.join(thisdir, 'testdata', 'cmake-format.py')
    self.tempconfig = os.path.join(self.tempdir, '.cmake-format.py')
    shutil.copyfile(configpath, self.tempconfig)

  @contextlib.contextmanager
  def subTest(self, msg=None, **params):
    # pylint: disable=no-member
    if sys.version_info < (3, 4, 0):
      yield None
    else:
      yield super(TestInvocations, self).subTest(msg=msg, **params)

  def tearDown(self):
    shutil.rmtree(self.tempdir)

  def test_pipeout_invocation(self):
    """
    Test invocation with an infile path and output to stdout.
    """

    thisdir = os.path.realpath(os.path.dirname(__file__))
    infile_path = os.path.join(thisdir, 'testdata', 'test_in.cmake')
    expectfile_path = os.path.join(thisdir, 'testdata', 'test_out.cmake')

    with io.open(os.path.join(self.tempdir, 'test_out.cmake'), 'wb') as outfile:
      subprocess.check_call([sys.executable, '-Bm', 'cmakelang.format',
                             infile_path], stdout=outfile, cwd=self.tempdir,
                            env=self.env)

    with io.open(os.path.join(self.tempdir, 'test_out.cmake'), 'r',
                 encoding='utf8') as infile:
      actual_text = infile.read()
    with io.open(expectfile_path, 'r', encoding='utf8') as infile:
      expected_text = infile.read()

    delta_lines = list(difflib.unified_diff(expected_text.split('\n'),
                                            actual_text.split('\n')))
    if delta_lines:
      raise AssertionError('\n'.join(delta_lines[2:]))

  def test_fileout_invocation(self):
    """
    Test invocation with an infile path and outfile path
    """

    thisdir = os.path.realpath(os.path.dirname(__file__))
    infile_path = os.path.join(thisdir, 'testdata', 'test_in.cmake')
    expectfile_path = os.path.join(thisdir, 'testdata', 'test_out.cmake')

    subprocess.check_call([sys.executable, '-Bm', 'cmakelang.format',
                           '-o', os.path.join(self.tempdir, 'test_out.cmake'),
                           infile_path], cwd=self.tempdir, env=self.env)

    with io.open(os.path.join(self.tempdir, 'test_out.cmake'), 'r',
                 encoding='utf8') as infile:
      actual_text = infile.read()
    with io.open(expectfile_path, 'r', encoding='utf8') as infile:
      expected_text = infile.read()

    delta_lines = list(difflib.unified_diff(expected_text.split('\n'),
                                            actual_text.split('\n')))
    if delta_lines:
      raise AssertionError('\n'.join(delta_lines[2:]))

  def test_inplace_invocation(self):
    """
    Test invocation for inplace format of a file
    """

    thisdir = os.path.realpath(os.path.dirname(__file__))
    infile_path = os.path.join(thisdir, 'testdata', 'test_in.cmake')
    expectfile_path = os.path.join(thisdir, 'testdata', 'test_out.cmake')

    ofd, tmpfile_path = tempfile.mkstemp(suffix='.txt', prefix='CMakeLists',
                                         dir=self.tempdir)
    os.close(ofd)
    shutil.copyfile(infile_path, tmpfile_path)
    os.chmod(tmpfile_path, 0o755)
    subprocess.check_call([sys.executable, '-Bm', 'cmakelang.format',
                           '-i', tmpfile_path], cwd=self.tempdir, env=self.env)

    with io.open(os.path.join(tmpfile_path), 'r', encoding='utf8') as infile:
      actual_text = infile.read()
    with io.open(expectfile_path, 'r', encoding='utf8') as infile:
      expected_text = infile.read()

    delta_lines = list(difflib.unified_diff(expected_text.split('\n'),
                                            actual_text.split('\n')))
    if delta_lines:
      raise AssertionError('\n'.join(delta_lines[2:]))

    # Verify permissions are preserved
    self.assertEqual(oct(os.stat(tmpfile_path).st_mode)[-3:], "755")

  def test_check_invocation(self):
    """
    Test invocation for --check of a file
    """

    thisdir = os.path.realpath(os.path.dirname(__file__))
    unformatted_path = os.path.join(thisdir, 'testdata', 'test_in.cmake')
    formatted_path = os.path.join(thisdir, 'testdata', 'test_out.cmake')

    with open("/dev/null", "wb") as devnull:
      statuscode = subprocess.call([
          sys.executable, '-Bm', 'cmakelang.format',
          '--check', unformatted_path], env=self.env, stderr=devnull)
    self.assertEqual(1, statuscode)

    statuscode = subprocess.call([
        sys.executable, '-Bm', 'cmakelang.format',
        '--check', formatted_path], env=self.env)
    self.assertEqual(0, statuscode)

  def test_stream_invocation(self):
    """
    Test invocation with stdin as the infile and stdout as the outifle
    """

    thisdir = os.path.realpath(os.path.dirname(__file__))
    infile_path = os.path.join(thisdir, 'testdata', 'test_in.cmake')
    expectfile_path = os.path.join(thisdir, 'testdata', 'test_out.cmake')

    stdinpipe = os.pipe()
    stdoutpipe = os.pipe()

    def preexec():
      os.close(stdinpipe[1])
      os.close(stdoutpipe[0])

    # pylint: disable=W1509
    proc = subprocess.Popen([sys.executable, '-Bm', 'cmakelang.format', '-'],
                            stdin=stdinpipe[0], stdout=stdoutpipe[1],
                            cwd=self.tempdir, env=self.env, preexec_fn=preexec)
    os.close(stdinpipe[0])
    os.close(stdoutpipe[1])

    with io.open(infile_path, 'r', encoding='utf-8') as infile:
      with io.open(stdinpipe[1], 'w', encoding='utf-8') as outfile:
        for line in infile:
          outfile.write(line)

    with io.open(stdoutpipe[0], 'r', encoding='utf-8') as infile:
      actual_text = infile.read()

    proc.wait()

    with io.open(expectfile_path, 'r', encoding='utf8') as infile:
      expected_text = infile.read()

    delta_lines = list(difflib.unified_diff(expected_text.split('\n'),
                                            actual_text.split('\n')))
    if delta_lines:
      raise AssertionError('\n'.join(delta_lines[2:]))

  def test_encoding_invocation(self):
    """
    Try to reformat latin1-encoded file, once with default
    encoding (-> prompt utf8-decoding error) and once with
    specifically latin1 encoding (-> should succeed)
    """

    thisdir = os.path.realpath(os.path.dirname(__file__))
    infile_path = os.path.join(thisdir, 'testdata', 'test_latin1_in.cmake')
    expectfile_path = os.path.join(thisdir, 'testdata', 'test_latin1_out.cmake')

    # this invocation should fail
    invocation_result = subprocess.call(
        [sys.executable, '-Bm', 'cmakelang.format',
         '--outfile-path', os.path.join(self.tempdir, 'test_latin1_out.cmake'),
         infile_path],
        cwd=self.tempdir, env=self.env,
        stderr=subprocess.PIPE)
    self.assertNotEqual(
        0, invocation_result,
        msg="Expected cmake-format invocation to fail but did not")

    # this invocation should succeed
    subprocess.check_call(
        [sys.executable, '-Bm', 'cmakelang.format',
         '--input-encoding=latin1',
         '--output-encoding=latin1',
         '--outfile-path', os.path.join(self.tempdir, 'test_latin1_out.cmake'),
         infile_path],
        cwd=self.tempdir, env=self.env)

    with io.open(os.path.join(self.tempdir, 'test_latin1_out.cmake'), 'r',
                 encoding='latin1') as infile:
      actual_text = infile.read()

    with io.open(expectfile_path, 'r', encoding='latin1') as infile:
      expected_text = infile.read()

    delta_lines = list(difflib.unified_diff(expected_text.split('\n'),
                                            actual_text.split('\n')))
    if delta_lines:
      raise AssertionError('\n'.join(delta_lines[2:]))

  def test_no_config_invocation(self):
    """
    Test invocation with no config file specified
    """
    os.unlink(self.tempconfig)
    thisdir = os.path.realpath(os.path.dirname(__file__))
    infile_path = os.path.join(thisdir, 'testdata', 'test_in.cmake')
    expectfile_path = os.path.join(thisdir, 'testdata', 'test_out.cmake')

    subprocess.check_call([sys.executable, '-Bm', 'cmakelang.format',
                           '-o', os.path.join(self.tempdir, 'test_out.cmake'),
                           infile_path], cwd=self.tempdir, env=self.env)

    with io.open(os.path.join(self.tempdir, 'test_out.cmake'), 'r',
                 encoding='utf8') as infile:
      actual_text = infile.read()
    with io.open(expectfile_path, 'r', encoding='utf8') as infile:
      expected_text = infile.read()

    delta_lines = list(difflib.unified_diff(expected_text.split('\n'),
                                            actual_text.split('\n')))
    if delta_lines:
      raise AssertionError('\n'.join(delta_lines[2:]))

  def test_multiple_config_invocation(self):
    """
    Repeat the default config test using a config file that is split. The
    custom command definitions are in the second file.
    """
    thisdir = os.path.realpath(os.path.dirname(__file__))
    configdir = os.path.join(thisdir, 'testdata')
    infile_path = os.path.join(thisdir, 'testdata', 'test_in.cmake')
    expectfile_path = os.path.join(thisdir, 'testdata', 'test_out.cmake')

    subprocess.check_call([
        sys.executable, '-Bm', 'cmakelang.format', '-c',
        os.path.join(configdir, 'cmake-format-split-1.py'),
        os.path.join(configdir, 'cmake-format-split-2.py'),
        '-o', os.path.join(self.tempdir, 'test_out.cmake'),
        infile_path], cwd=self.tempdir, env=self.env)

    with io.open(os.path.join(self.tempdir, 'test_out.cmake'), 'r',
                 encoding='utf8') as infile:
      actual_text = infile.read()
    with io.open(expectfile_path, 'r', encoding='utf8') as infile:
      expected_text = infile.read()

    delta_lines = list(difflib.unified_diff(expected_text.split('\n'),
                                            actual_text.split('\n')))
    if delta_lines:
      raise AssertionError('\n'.join(delta_lines[2:]))

  def test_multiple_config_invocation_dupflag(self):
    """
    Repeat the default config test using a config file that is split. The
    custom command definitions are in the second file.
    """
    thisdir = os.path.realpath(os.path.dirname(__file__))
    configdir = os.path.join(thisdir, 'testdata')
    infile_path = os.path.join(thisdir, 'testdata', 'test_in.cmake')
    expectfile_path = os.path.join(thisdir, 'testdata', 'test_out.cmake')

    subprocess.check_call([
        sys.executable, '-Bm', 'cmakelang.format',
        '-c', os.path.join(configdir, 'cmake-format-split-1.py'),
        '-c', os.path.join(configdir, 'cmake-format-split-2.py'),
        '-o', os.path.join(self.tempdir, 'test_out.cmake'),
        infile_path], cwd=self.tempdir, env=self.env)

    with io.open(os.path.join(self.tempdir, 'test_out.cmake'), 'r',
                 encoding='utf8') as infile:
      actual_text = infile.read()
    with io.open(expectfile_path, 'r', encoding='utf8') as infile:
      expected_text = infile.read()

    delta_lines = list(difflib.unified_diff(expected_text.split('\n'),
                                            actual_text.split('\n')))
    if delta_lines:
      raise AssertionError('\n'.join(delta_lines[2:]))

  def test_auto_lineendings(self):
    """
    Verify that windows line-endings are detected and preserved on input.
    """
    thisdir = os.path.realpath(os.path.dirname(__file__))
    for suffix in ["win", "unix"]:
      with self.subTest(lineending=suffix):
        filename = "test_lineend_{}.cmake".format(suffix)
        infile_path = os.path.join(thisdir, 'testdata', filename)
        outfile_path = os.path.join(self.tempdir, filename)
        subprocess.check_call([sys.executable, '-Bm', 'cmakelang.format',
                               '--line-ending=auto',
                               '-o', outfile_path, infile_path],
                              cwd=self.tempdir, env=self.env)
        with open(infile_path, "rb") as infile:
          infile_bytes = infile.read()
        with open(outfile_path, "rb") as infile:
          outfile_bytes = infile.read()
        self.assertEqual(infile_bytes, outfile_bytes)

  def test_notouch(self):
    """
    Verify that, if formatting is unchanged, an --in-place file is not modified
    """
    thisdir = os.path.realpath(os.path.dirname(__file__))
    expectfile_path = os.path.join(thisdir, 'testdata', 'test_out.cmake')
    outfile_path = os.path.join(self.tempdir, 'test_out.cmake')
    shutil.copy2(expectfile_path, outfile_path)
    mtime_before = os.path.getmtime(outfile_path)
    subprocess.check_call(
        [sys.executable, '-Bm', 'cmakelang.format', '-i', outfile_path],
        cwd=self.tempdir, env=self.env)
    mtime_after = os.path.getmtime(outfile_path)
    self.assertEqual(mtime_before, mtime_after)

  def test_require_valid(self):
    """
    Verify that the --require-valid-layout flag works as intended
    """
    thisdir = os.path.realpath(os.path.dirname(__file__))
    testfilepath = os.path.join(thisdir, 'testdata', 'test_invalid.cmake')

    with tempfile.NamedTemporaryFile(
        suffix=".txt", prefix="CMakeLists", dir=self.tempdir) as outfile:
      statuscode = subprocess.call(
          [sys.executable, '-Bm', 'cmakelang.format', testfilepath],
          stdout=outfile, stderr=outfile, env=self.env)
    self.assertEqual(0, statuscode)

    with tempfile.NamedTemporaryFile(
        suffix=".txt", prefix="CMakeLists", dir=self.tempdir) as outfile:
      statuscode = subprocess.call(
          [sys.executable, '-Bm', 'cmakelang.format', testfilepath,
           "--require-valid-layout"],
          stdout=outfile, stderr=outfile, env=self.env)
    self.assertEqual(1, statuscode)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""
Generate linter documentation page
"""

from __future__ import unicode_literals

import argparse
import io
import logging
import os
import sys
import textwrap

import cmakelang
from cmakelang.lint import lintdb
from tangent.tooling.gendoc import format_directive

HEADER = """
.. _lint-checks:

===================
Lint Code Reference
===================

"""


def setup_argparse(argparser):
  argparser.add_argument('-v', '--version', action='version',
                         version=cmakelang.__version__)
  argparser.add_argument(
      '-l', '--log-level', default="info",
      choices=["error", "warning", "info", "debug"])

  argparser.add_argument(
      '-o', '--outfile-path', default=None,
      help='Write output to this file. Default is stdout.')

  argparser.add_argument(
      '-c', '--config-files', nargs='+',
      help='path to configuration file(s)')
  argparser.add_argument('infilepaths', nargs='*')
  argparser.add_argument("command", choices=["reference", "table"])


def write_title(outfile, title, rulerchar=None, numrule=1):
  if rulerchar is None:
    rulerchar = '-'

  if numrule == 2:
    outfile.write(rulerchar * len(title))
    outfile.write("\n")
  outfile.write(title)
  outfile.write("\n")
  outfile.write(rulerchar * len(title))
  outfile.write("\n\n")


def write_ruler(outfile):
  outfile.write("+")
  outfile.write("-" * 7)
  outfile.write("+")
  outfile.write("-" * 68)
  outfile.write("+")
  outfile.write("\n")


def write_cell(outfile, idstr, msgfmt):
  lines = textwrap.wrap(msgfmt, width=66)
  outfile.write("|`{:5s}`| {:66s} |\n".format(idstr, lines.pop(0)))
  for line in lines:
    outfile.write("| {:5s} | {:66s} |\n".format("", line))


def gen_table(outfile):

  write_ruler(outfile)
  for idstr, msgfmt, _ in lintdb.LINT_DB:
    write_cell(outfile, idstr, msgfmt)
    write_ruler(outfile)
  outfile.write("\n")


def gen_reference(outfile):
  outfile.write(HEADER)
  for idstr, msgfmt, kwargs in lintdb.LINT_DB:
    # outfile.write(".. _{}:\n".format(idstr.lower()))
    outfile.write(".. _{}:\n\n".format(idstr))
    write_title(outfile, idstr, numrule=2)
    write_title(outfile, "message")
    outfile.write(format_directive(msgfmt))
    outfile.write("\n\n")
    description = kwargs.pop("description", None)
    if description:
      write_title(outfile, "description")
      outfile.write(description)
      outfile.write("\n\n")

    explain = kwargs.pop("explain", None)
    if explain:
      write_title(outfile, "explanation")
      outfile.write(explain)
      outfile.write("\n\n")


def main():
  """Parse arguments, open files, start work."""

  argparser = argparse.ArgumentParser(
      description=__doc__,
      formatter_class=argparse.RawDescriptionHelpFormatter)

  setup_argparse(argparser)
  args = argparser.parse_args()
  logging.getLogger().setLevel(getattr(logging, args.log_level.upper()))

  if args.outfile_path is None:
    args.outfile_path = '-'

  if args.outfile_path == '-':
    outfile = io.open(os.dup(sys.stdout.fileno()),
                      mode='w', encoding="utf-8", newline='')
  else:
    outfile = io.open(args.outfile_path, 'w', encoding="utf-8", newline='')

  if args.command == "reference":
    gen_reference(outfile)
  elif args.command == "table":
    gen_table(outfile)
  outfile.close()
  return 0


if __name__ == "__main__":
  logging.basicConfig(level=logging.INFO, format="%(levelname)s %(message)s")
  sys.exit(main())



import logging

from cmakelang.lex import TokenType
from cmakelang.parse.additional_nodes import PatternNode
from cmakelang.parse.argument_nodes import (
    ArgGroupNode,
    ConditionalGroupNode,
    PositionalGroupNode,
    PositionalParser,
    StandardArgTree,
    StandardParser)
from cmakelang.parse.util import get_first_semantic_token
from cmakelang.parse.simple_nodes import consume_whitespace_and_comments

logger = logging.getLogger(__name__)


def parse_file_read(ctx, tokens, breakstack):
  """
  ::

    file(READ <filename> <variable>
         [OFFSET <offset>] [LIMIT <max-in>] [HEX])

  :see: https://cmake.org/cmake/help/v3.14/command/file.html#read
  """

  return StandardArgTree.parse(
      ctx, tokens,
      npargs='*',
      kwargs={
          "OFFSET": PositionalParser(1),
          "LIMIT": PositionalParser(1),
      },
      flags=[
          "READ",
          "HEX"
      ],
      breakstack=breakstack)


def parse_file_strings(ctx, tokens, breakstack):
  """
  ::

    file(STRINGS <filename> <variable> [<options>...])

  :see: https://cmake.org/cmake/help/v3.14/command/file.html#read
  """

  return StandardArgTree.parse(
      ctx, tokens,
      npargs='*',
      kwargs={
          "LENGTH_MAXIMUM": PositionalParser(1),
          "LENGTH_MINIMUM": PositionalParser(1),
          "LIMIT_COUNT": PositionalParser(1),
          "LIMIT_INPUT": PositionalParser(1),
          "LIMIT_OUTPUT": PositionalParser(1),
          "REGEX": PositionalParser(1),
          "ENCODING": PositionalParser(1, flags=[
              "UTF-8", "UTF-16LE", "UTF-16BE", "UTF-32LE", "UTF-32BE"
          ]),
      },
      flags=[
          "STRINGS",
          "NEWLINE_CONSUME",
          "NO_HEX_CONVERSION",

      ],
      breakstack=breakstack)


HASH_STRINGS = [
    "MD5",
    "SHA1",
    "SHA224",
    "SHA256",
    "SHA384",
    "SHA512",
    "SHA3_224",
    "SHA3_256",
    "SHA3_384",
    "SHA3_512",
]


def parse_file_hash(ctx, tokens, breakstack):
  """
  ::

    file(<HASH> <filename> <variable>)

  :see: https://cmake.org/cmake/help/v3.14/command/file.html#strings
  """

  return StandardArgTree.parse(
      ctx, tokens,
      npargs=3,
      kwargs={},
      flags=HASH_STRINGS,
      breakstack=breakstack)


def parse_file_timestamp(ctx, tokens, breakstack):
  """
  ::

    file(TIMESTAMP <filename> <variable> [<format>] [UTC])

  :see: https://cmake.org/cmake/help/v3.14/command/file.html#strings
  """

  return StandardArgTree.parse(
      ctx, tokens,
      npargs='+',
      kwargs={},
      flags=["TIMESTAMP", "UTC"],
      breakstack=breakstack)


def parse_file_write(ctx, tokens, breakstack):
  """
  ::

    file(WRITE <filename> <content>...)
    file(APPEND <filename> <content>...)

  :see: https://cmake.org/cmake/help/v3.14/command/file.html#writing
  """
  tree = ArgGroupNode()
  consume_whitespace_and_comments(ctx, tokens, tree)
  tree.children.append(
      PositionalGroupNode.parse(
          ctx, tokens, 2, ["WRITE", "APPEND"], breakstack))
  consume_whitespace_and_comments(ctx, tokens, tree)
  tree.children.append(
      PositionalGroupNode.parse(ctx, tokens, '+', [], breakstack))

  return tree


def parse_file_generate_output(ctx, tokens, breakstack):
  """
  ::

    file(GENERATE OUTPUT output-file
        <INPUT input-file|CONTENT content>
        [CONDITION expression])

  :see: https://cmake.org/cmake/help/v3.14/command/file.html#writing
  """

  return StandardArgTree.parse(
      ctx, tokens,
      npargs=1,
      kwargs={
          "OUTPUT": PositionalParser(1),
          "INPUT": PositionalParser(1),
          "CONTENT": PositionalParser('+'),
          "CONDITION": ConditionalGroupNode.parse,
      },
      flags=["GENERATE"],
      breakstack=breakstack)


def parse_file_glob(ctx, tokens, breakstack):
  """
  ::

    file(GLOB <variable>
        [LIST_DIRECTORIES true|false] [RELATIVE <path>] [CONFIGURE_DEPENDS]
        [<globbing-expressions>...])
    file(GLOB_RECURSE <variable> [FOLLOW_SYMLINKS]
        [LIST_DIRECTORIES true|false] [RELATIVE <path>] [CONFIGURE_DEPENDS]
        [<globbing-expressions>...])
  :see: https://cmake.org/cmake/help/v3.14/command/file.html#filesystem
  """

  return StandardArgTree.parse(
      ctx, tokens,
      npargs='+',
      kwargs={
          "LIST_DIRECTORIES": PositionalParser(1),
          "RELATIVE": PositionalParser(1)
      },
      flags=[
          "GLOB",
          "GLOB_RECURSE",
          "CONFIGURE_DEPENDS",
          "FOLLOW_SYMLINKS"],
      breakstack=breakstack)


def parse_file_copy(ctx, tokens, breakstack):
  """
  ::

    file(<COPY|INSTALL> <files>... DESTINATION <dir>
         [FILE_PERMISSIONS <permissions>...]
         [DIRECTORY_PERMISSIONS <permissions>...]
         [NO_SOURCE_PERMISSIONS] [USE_SOURCE_PERMISSIONS]
         [FILES_MATCHING]
         [[PATTERN <pattern> | REGEX <regex>]
          [EXCLUDE] [PERMISSIONS <permissions>...]] [...])

  :see: https://cmake.org/cmake/help/v3.14/command/file.html#filesystem
  """

  return StandardArgTree.parse(
      ctx, tokens,
      npargs='*',
      kwargs={
          "COPY": PositionalParser('*'),
          "DESTINATION": PositionalParser(1),
          "FILE_PERMISSIONS":
              PositionalParser(
                  '+',
                  flags=["OWNER_READ", "OWNER_WRITE", "OWNER_EXECUTE",
                         "GROUP_READ", "GROUP_WRITE", "GROUP_EXECUTE",
                         "WORLD_READ", "WORLD_WRITE", "WORLD_EXECUTE",
                         "SETUID", "SETGID"]),
          "DIRECTORY_PERMISSIONS": PositionalParser('+'),
          "PATTERN": PatternNode.parse,
          "REGEX": PatternNode.parse,
      },
      flags=[
          "INSTALL",
          "NO_SOURCE_PERMISSIONS",
          "USE_SOURCE_PERMISSIONS",
          "FILES_MATCHING",
      ],
      breakstack=breakstack)


def parse_file_create_link(ctx, tokens, breakstack):
  """
  ::

    file(CREATE_LINK <original> <linkname>
        [RESULT <result>] [COPY_ON_ERROR] [SYMBOLIC])

  :see: https://cmake.org/cmake/help/v3.14/command/file.html#filesystem
  """

  return StandardArgTree.parse(
      ctx, tokens,
      npargs='+',
      kwargs={
          "RESULT": PositionalParser(1),
      },
      flags=[
          "COPY_ON_ERROR",
          "SYMBOLIC",
      ],
      breakstack=breakstack)


def parse_file_xfer(ctx, tokens, breakstack):
  """
  ::

    file(DOWNLOAD <url> <file> [<options>...])
    file(UPLOAD   <file> <url> [<options>...])

  :see: https://cmake.org/cmake/help/v3.14/command/file.html#transfer
  """

  return StandardArgTree.parse(
      ctx, tokens,
      npargs=3,
      kwargs={
          "INACTIVITY_TIMEOUT": PositionalParser(1),
          "LOG": PositionalParser(1),
          "STATUS": PositionalParser(1),
          "TIMEOUT": PositionalParser(1),
          "USERPWD": PositionalParser(1),
          "HTTPHEADER": PositionalParser(1),
          "NETRC": PositionalParser(
              1, flags=["CMAKE_NETRC", "IGNORED", "OPTIONAL", "REQUIRED"]),
          "NETRC_FILE": PositionalParser(1),
          "EXPECTED_HASH": PositionalParser(1),
          "EXPECTED_MD5": PositionalParser(1),
          "TLS_VERIFY": PositionalParser(1),
          "TLS_CAINFO": PositionalParser(1),
      },
      flags=[
          "DOWNLOAD",
          "UPLOAD",
          "SHOW_PROGRESS",

      ],
      breakstack=breakstack)


def parse_file_lock(ctx, tokens, breakstack):
  """
  ::

    file(LOCK <path> [DIRECTORY] [RELEASE]
         [GUARD <FUNCTION|FILE|PROCESS>]
         [RESULT_VARIABLE <variable>]
         [TIMEOUT <seconds>])

  :see: https://cmake.org/cmake/help/v3.14/command/file.html#locking
  """

  return StandardArgTree.parse(
      ctx, tokens,
      npargs='+',
      kwargs={
          "GUARD": PositionalParser(1, flags=["FUNCTION", "FILE", "PROCESS"]),
          "RESULT_VARIABLE": PositionalParser(1),
          "TIMEOUT": PositionalParser(1)
      },
      flags=[
          "LOCK",
          "DIRECTORY",
          "RELEASE",
      ],
      breakstack=breakstack)


def parse_file(ctx, tokens, breakstack):
  """
  The ``file()`` command has a lot of different forms, depending on the first
  argument. This function just dispatches the correct parse implementation for
  the given form::

    Reading
      file(READ <filename> <out-var> [...])
      file(STRINGS <filename> <out-var> [...])
      file(<HASH> <filename> <out-var>)
      file(TIMESTAMP <filename> <out-var> [...])

    Writing
      file({WRITE | APPEND} <filename> <content>...)
      file({TOUCH | TOUCH_NOCREATE} [<file>...])
      file(GENERATE OUTPUT <output-file> [...])

    Filesystem
      file({GLOB | GLOB_RECURSE} <out-var> [...] [<globbing-expr>...])
      file(RENAME <oldname> <newname>)
      file({REMOVE | REMOVE_RECURSE } [<files>...])
      file(MAKE_DIRECTORY [<dir>...])
      file({COPY | INSTALL} <file>... DESTINATION <dir> [...])
      file(SIZE <filename> <out-var>)
      file(READ_SYMLINK <linkname> <out-var>)
      file(CREATE_LINK <original> <linkname> [...])

    Path Conversion
      file(RELATIVE_PATH <out-var> <directory> <file>)
      file({TO_CMAKE_PATH | TO_NATIVE_PATH} <path> <out-var>)

    Transfer
      file(DOWNLOAD <url> <file> [...])
      file(UPLOAD <file> <url> [...])

    Locking
      file(LOCK <path> [...])

  :see: https://cmake.org/cmake/help/v3.14/command/file.html
  """

  descriminator_token = get_first_semantic_token(tokens)
  if (descriminator_token is None or
      descriminator_token.type is TokenType.RIGHT_PAREN):
    location = ()
    if tokens:
      location = tokens[0].get_location()
    # logger.warning("Invalid empty file() command at %s", location)
    ctx.lint_ctx.record_lint("E1120", location=location)
    return StandardArgTree.parse(ctx, tokens, npargs='*', kwargs={}, flags=[],
                                 breakstack=breakstack)

  if descriminator_token.type is TokenType.DEREF:
    ctx.lint_ctx.record_lint(
        "C0114", location=descriminator_token.get_location())
    return StandardArgTree.parse(ctx, tokens, npargs='*', kwargs={}, flags=[],
                                 breakstack=breakstack)

  descriminator = descriminator_token.spelling.upper()
  parsemap = {
      "READ": parse_file_read,
      "STRINGS": parse_file_strings,
      "TIMESTAMP": parse_file_timestamp,
      "WRITE": parse_file_write,
      "APPEND": parse_file_write,
      "TOUCH": StandardParser('+', flags=["TOUCH"]),
      "TOUCH_NOCREATE": StandardParser('+', flags=["TOUCH_NOCREATE"]),
      "GENERATE": parse_file_generate_output,
      "GLOB": parse_file_glob,
      "GLOB_RECURSE": parse_file_glob,
      "RENAME": StandardParser(3, flags=["RENAME"]),
      "REMOVE": StandardParser('+', flags=["REMOVE"]),
      "REMOVE_RECURSE": StandardParser('+', flags=["REMOVE_RECURSE"]),
      "MAKE_DIRECTORY": StandardParser('+', flags=["MAKE_DIRECTORY"]),
      "COPY": parse_file_copy,
      "INSTALL": parse_file_copy,
      "SIZE": StandardParser(3, flags=["SIZE"]),
      "READ_SYMLINK": StandardParser(3, flags=["READ_SYMLINK"]),
      "CREATE_LINK": parse_file_create_link,
      "RELATIVE_PATH": StandardParser(4, flags=["RELATIVE_PATH"]),
      "TO_CMAKE_PATH": StandardParser(3, flags=["TO_CMAKE_PATH"]),
      "TO_NATIVE_PATH": StandardParser(3, flags=["TO_NATIVE_PATH"]),
      "DOWNLOAD": parse_file_xfer,
      "UPLOAD": parse_file_xfer,
      "LOCK": parse_file_lock
  }
  for hashname in HASH_STRINGS:
    parsemap[hashname] = parse_file_hash

  if descriminator not in parsemap:
    # logger.warning(
    #     "Indeterminate form of file() command \"%s\" at %s", descriminator,
    #     tokens[0].location())
    ctx.lint_ctx.record_lint(
        "E1126", location=descriminator_token.get_location())
    return StandardArgTree.parse(ctx, tokens, npargs='*', kwargs={}, flags=[],
                                 breakstack=breakstack)

  return parsemap[descriminator](ctx, tokens, breakstack)


def populate_db(parse_db):
  parse_db["file"] = parse_file



from __future__ import print_function

from cmakelang import lex


def main():
  """
  Dump tokenized listfile to stdout for debugging.
  """
  import argparse
  parser = argparse.ArgumentParser(description=__doc__)
  parser.add_argument('infile')
  args = parser.parse_args()
  with open(args.infile, 'r') as infile:
    tokens = lex.tokenize(infile.read())
  print('\n'.join(str(x) for x in tokens))


if __name__ == '__main__':
  main()



"""
Some of the test lint would get removed by my editor we have a little script to
help generate the file.
"""

import io
import os
import re
import logging

logger = logging.getLogger(__name__)


def make_expect_lint():
  thisdir = os.path.dirname(os.path.realpath(__file__))

  # strip test annotations from the test file make the documentation a little
  # cleaner
  with io.open(os.path.join(thisdir, "lint_tests.cmake"), newline='') as infile:
    content = infile.read()
  content = re.sub(
      "^# ((test)|(expect)): .*\n", "", content, flags=re.MULTILINE)
  with io.open(
      os.path.join(thisdir, "expect_lint.cmake"), "w", newline='') as outfile:
    outfile.write(content)


def rewrite_lint_tests():
  thisdir = os.path.dirname(os.path.realpath(__file__))

  # re-process the test file to re-introduce lint that is likely to have been
  # removed by an editor
  with io.open(os.path.join(thisdir, "lint_tests.cmake"), newline='') as infile:
    content = infile.read()

  lines = content.split("\n")
  have_changes = False

  # Need to remove newline
  if content[-1] == "\n":
    logger.info("Removing final newline")
    have_changes = True

  outlines = []
  for line in lines:
    if line.rstrip() == "# This line has the wrong line endings":
      if not line.endswith("\r"):
        logger.info("Adding carriage return to line ending test")
        line = line.rstrip() + "\r"
        have_changes = True

    if line.rstrip() == "# This line has trailing whitespace":
      if not line.endswith(" "):
        logger.info("Adding whitespace to trailing whitespace test")
        line = line.rstrip() + "   "
        have_changes = True
    outlines.append(line)

  if have_changes:
    while outlines and not outlines[-1].strip():
      outlines.pop(-1)
    with io.open(os.path.join(
        thisdir, "lint_tests.cmake"), "w", newline='') as outfile:
      outfile.write("\n".join(outlines))


def genfiles():
  rewrite_lint_tests()
  make_expect_lint()


if __name__ == "__main__":
  logging.basicConfig(level=logging.INFO)
  genfiles()



# -*- coding: utf-8 -*-
# pylint: disable=W0613
from __future__ import print_function
from __future__ import unicode_literals

import logging

from cmakelang import lex
from cmakelang.common import UserError
from cmakelang.parse.util import COMMENT_TOKENS, WHITESPACE_TOKENS
from cmakelang.parse.common import NodeType, ParenBreaker, TreeNode
from cmakelang.parse.printer import tree_string
from cmakelang.parse.argument_nodes import StandardParser, StandardParser2
from cmakelang.parse.simple_nodes import CommentNode


logger = logging.getLogger(__name__)


class FunctionNameNode(TreeNode):
  def __init__(self):
    super(FunctionNameNode, self).__init__(NodeType.FUNNAME)
    self.token = None

  @classmethod
  def parse(cls, ctx, tokens):
    node = cls()
    node.token = tokens.pop(0)
    node.children.append(node.token)
    return node


class StatementNode(TreeNode):
  """Parent node of a statement subtree."""

  def __init__(self):
    super(StatementNode, self).__init__(NodeType.STATEMENT)
    self.funnode = None
    self.argtree = None
    self.cmdspec = None

  def get_funname(self):
    return self.funnode.token.spelling.lower()

  @classmethod
  def consume(cls, ctx, tokens):
    """
    Consume a complete statement, removing tokens from the input list and
    returning a STATEMENT node.
    """
    node = cls()

    # Consume the function name
    fnname = tokens[0].spelling.lower()
    node.funnode = funnode = FunctionNameNode.parse(ctx, tokens)
    node.children.append(funnode)

    # Consume whitespace up to the parenthesis
    while tokens and tokens[0].type in WHITESPACE_TOKENS:
      node.children.append(tokens.pop(0))

    # TODO(josh): should the parens belong to the statement node or the
    # group node?
    if tokens[0].type != lex.TokenType.LEFT_PAREN:
      raise ValueError(
          "Unexpected {} token at {}, expecting l-paren, got {}"
          .format(tokens[0].type.name, tokens[0].get_location(),
                  repr(tokens[0].content)))

    lparen = TreeNode(NodeType.LPAREN)
    lparen.children.append(tokens.pop(0))
    node.children.append(lparen)

    while tokens and tokens[0].type in WHITESPACE_TOKENS:
      node.children.append(tokens.pop(0))
      continue

    breakstack = [ParenBreaker()]

    parse_fun = ctx.parse_db.get(fnname, None)
    if parse_fun is None:
      # If the parse_db provides a "_default" then use that. Otherwise use the
      # standard parser with no kwargs or flags.
      parse_fun = ctx.parse_db.get("_default", StandardParser())

    if isinstance(parse_fun, StandardParser2):
      node.cmdspec = parse_fun.cmdspec

    node.argtree = subtree = parse_fun(ctx, tokens, breakstack)
    node.children.append(subtree)

    # NOTE(josh): technically we may have a statement specification with
    # an exact number of arguments. At this point we have broken out of that
    # statement but we might have some comments or whitespace to consume
    while tokens and tokens[0].type != lex.TokenType.RIGHT_PAREN:
      if tokens[0].type in WHITESPACE_TOKENS:
        node.children.append(tokens.pop(0))
        continue

      if tokens[0].type in COMMENT_TOKENS:
        cnode = CommentNode.consume(ctx, tokens)
        node.children.append(cnode)
        continue

      raise UserError(
          "Unexpected {} token at {}, expecting r-paren, got {}"
          .format(tokens[0].type.name, tokens[0].get_location(),
                  repr(tokens[0].content)))

    if not tokens:
      raise UserError(
          "Unexpected end of token stream while parsing statement:\n {}"
          .format(tree_string([node])))

    if tokens[0].type != lex.TokenType.RIGHT_PAREN:
      raise UserError(
          "Unexpected {} token at {}, expecting r-paren, got {}"
          .format(tokens[0].type.name, tokens[0].get_location(),
                  repr(tokens[0].content)))

    rparen = TreeNode(NodeType.RPAREN)
    rparen.children.append(tokens.pop(0))
    node.children.append(rparen)
    CommentNode.consume_trailing(ctx, tokens, node)

    return node


class AtWordNode(TreeNode):
  def __init__(self):
    super(AtWordNode, self).__init__(NodeType.ATWORD)
    self.token = None

  @classmethod
  def parse(cls, ctx, tokens):
    node = cls()
    node.token = tokens.pop(0)
    node.children.append(node.token)
    return node


class AtWordStatementNode(TreeNode):
  """Parent node of a statement subtree."""

  def __init__(self):
    super(AtWordStatementNode, self).__init__(NodeType.ATWORDSTATEMENT)
    self.atword_node = None

  @classmethod
  def consume(cls, ctx, tokens):
    """
    Consume an at-word substitution for a complete statement, removing tokens
    from the input list and returning an AtWordStatementNode.
    """
    node = cls()

    # Consume the replacement tokens (at-word)
    node.atword_node = AtWordNode.parse(ctx, tokens)
    node.children.append(node.atword_node)

    # Consume any trailing comments
    CommentNode.consume_trailing(ctx, tokens, node)
    return node



"""
Bump cmake version number
"""

import argparse
import collections
import io
import logging
import json
import os
import re
import sys

logger = logging.getLogger(__name__)


def stringify(version):
  """Given a version as a list/tuple of integers, return a list of strings
     coresponding to the different semantic version components."""
  version = list(str(x) for x in version)
  if len(version) > 3:
    version[3] = "dev" + version[3]
  return version


def format_pip_vstring(version):
  """Format the version items as a pip version string."""
  return ".".join(stringify(version))


def format_semver(version):
  """Format the version items as a semver."""
  version = stringify(version)
  left_part = ".".join(version[:3])
  right_part = ".".join(version[3:])
  parts = [left_part]
  if right_part:
    parts.append(right_part)
  return "-".join(parts)


INIT_VERSION_PATTERN = r"VERSION = '(\d+)\.(\d+)\.(\d+)(?:\.dev(\d+))'"


def get_current_version(filepath):
  """
  Process the __init__.py which is the canonical version number source.
  Return the current version as a list of integers.
  """
  pattern = re.compile(INIT_VERSION_PATTERN)
  with io.open(filepath, "r", encoding="utf-8") as infile:
    for line in infile:
      match = pattern.match(line.strip())
      if match:
        return [int(x) for x in match.groups()]
  raise RuntimeError("Failed to find the current version in __init__.py")


def process_init(filepath, version):
  """
  Process the __init__.py which is the canonical version number source.
  Increment the desired semver index by one. Write out the new __init__.py
  and return the semver tuple.
  """
  pattern = re.compile(INIT_VERSION_PATTERN)
  outpath = filepath + ".tmp"
  with io.open(filepath, "r", encoding="utf-8") as infile:
    with io.open(outpath, "w", encoding="utf-8") as outfile:
      for line in infile:
        match = pattern.match(line.strip())
        if match:
          line = "VERSION = '{}'\n".format(format_pip_vstring(version))
        outfile.write(line)
  os.rename(outpath, filepath)
  return version


def process_installation_rst(filepath, version):
  """
  Process doc/installation.rst which uses the version number in some examples
  """
  pattern = re.compile(r"v(\d+)\.(\d+)\.(\d+)")
  repl = "v{0}.{1}.{2}".format(*version[:3])
  outpath = filepath + ".tmp"

  with io.open(filepath, "r", encoding="utf-8") as infile:
    with io.open(outpath, "w", encoding="utf-8") as outfile:
      for line in infile:
        line = pattern.sub(repl, line)
        outfile.write(line)
  os.rename(outpath, filepath)


def process_json(filepath, version):
  """
  Process the vscode extension package.json and package-lock.json files
  """
  with io.open(filepath, "r", encoding="utf-8") as infile:
    data = json.load(infile, object_pairs_hook=collections.OrderedDict)
  data["version"] = format_semver(version)
  with io.open(filepath, "w", encoding="utf-8") as outfile:
    json.dump(data, outfile, indent=4)
    outfile.write("\n")


def main():
  fields = ["major", "minor", "patch", "dev", "drop-dev"]
  parser = argparse.ArgumentParser(description=__doc__)
  parser.add_argument("field", nargs="?", choices=fields, default="dev")
  args = parser.parse_args()

  thisdir = os.path.dirname(os.path.realpath(__file__))
  rootdir = os.path.dirname(os.path.dirname(thisdir))
  init_path = os.path.join(rootdir, "cmakelang/__init__.py")
  current_version = get_current_version(init_path)

  if args.field == "drop-dev":
    new_version = current_version[:3]
  else:
    field_idx = fields.index(args.field)
    new_version = list(current_version)
    new_version[field_idx] += 1
    for idx in range(field_idx + 1, len(new_version)):
      new_version[idx] = 0

  process_init(init_path, new_version)
  process_installation_rst(
      os.path.join(rootdir, "cmakelang/doc/installation.rst"), new_version)
  process_json(
      os.path.join(rootdir, "cmakelang/vscode_extension/package.json"),
      new_version)
  process_json(
      os.path.join(rootdir, "cmakelang/vscode_extension/package-lock.json"),
      new_version)
  return 0


if __name__ == "__main__":
  sys.exit(main())



# -*- coding: utf-8 -*-
# pylint: disable=too-many-lines
from __future__ import print_function
from __future__ import unicode_literals

import logging
from operator import itemgetter as _itemgetter
import re
import sys

from cmakelang import lex
from cmakelang.common import UserError, InternalError

logger = logging.getLogger(__name__)

ZERO_OR_MORE = '*'
ONE_OR_MORE = '+'

if sys.version_info[0] < 3:
  STRING_TYPES = (str, unicode)
else:
  STRING_TYPES = (str,)

IMPLICIT_PARG_TYPES = STRING_TYPES + (int,)

WHITESPACE_TOKENS = (lex.TokenType.WHITESPACE,
                     lex.TokenType.NEWLINE)

# TODO(josh): Don't include FORMAT_OFF and FORMAT_ON in comment tokens. They
# wont get reflowed within a comment block so there is no reason to parse them
# as such.
COMMENT_TOKENS = (lex.TokenType.COMMENT,)

ONOFF_TOKENS = (lex.TokenType.FORMAT_ON,
                lex.TokenType.FORMAT_OFF)

ALL_COMMENT_TOKENS = (
    lex.TokenType.COMMENT,
    lex.TokenType.BRACKET_COMMENT,
    lex.TokenType.FORMAT_ON,
    lex.TokenType.FORMAT_OFF
)

NON_SEMANTIC_TOKENS = ALL_COMMENT_TOKENS + WHITESPACE_TOKENS


class PositionalSpec(tuple):
  """
  Encapsulates the parse specification for a positional argument group.
  NOTE(josh): this is a named tuple with default arguments and some init
  processing. If it wasn't for the init processing, we could just do:

  PositionalSpec = collections.namedtuple(
    "PositionalSpec", ["nargs", ...])
  PositionalSpec.__new__.__defaults__ = (False, None, None, False)

  But we don't want to self.tags and self.flags to point to a mutable global
  variable...
  """

  def __new__(cls, nargs, sortable=False, tags=None, flags=None, legacy=False,
              max_pargs_hwrap=None, always_wrap=None):
    if not tags:
      tags = []
    if not flags:
      flags = []
    return tuple.__new__(
        cls, (nargs, sortable, tags, flags, legacy, max_pargs_hwrap,
              always_wrap))

  nargs = property(_itemgetter(0))
  npargs = property(_itemgetter(0))
  sortable = property(_itemgetter(1))
  tags = property(_itemgetter(2))
  flags = property(_itemgetter(3))
  legacy = property(_itemgetter(4))
  max_pargs_hwrap = property(_itemgetter(5))
  always_wrap = property(_itemgetter(6))

  def replace(self, **kwargs):
    selfdict = {
        "nargs": self.nargs,
        "sortable": self.sortable,
        "tags": list(self.tags),
        "flags": list(self.flags),
        "legacy": self.legacy,
        "max_pargs_hwrap": self.max_pargs_hwrap,
        "always_wrap": self.always_wrap
    }
    selfdict.update(kwargs)
    return PositionalSpec(**selfdict)


def is_whitespace_token(token):
  return token.type in WHITESPACE_TOKENS


def is_comment_token(token):
  return token.type in COMMENT_TOKENS


def is_syntactic_token(token):
  """Return true for everything that isn't whitespace"""
  if token.type in WHITESPACE_TOKENS:
    return False
  return True


def is_semantic_token(token):
  """Return true for everything that isnt' whitespace or a comment"""
  if token.type in NON_SEMANTIC_TOKENS:
    return False
  return True


def is_variable_dereference(token):
  return token.type is lex.TokenType.DEREF


def is_valid_trailing_comment(token):
  """
  Return true if the token is a valid trailing comment
  """
  return (token.type in (lex.TokenType.COMMENT,
                         lex.TokenType.BRACKET_COMMENT) and
          not comment_is_tag(token))


def is_comment_matching_pattern(token, regex):
  """
  Return true if the token is an explicit trailing comment
  """
  if token.type is not lex.TokenType.COMMENT:
    return False

  return bool(regex.match(token.spelling))


def next_is_explicit_trailing_comment(config, tokens):
  """
  Return true if the next comment is an explicit trailing comment, false
  otherwise
  """

  regex = re.compile("^" + config.markup.explicit_trailing_pattern + ".*")
  for token in iter_syntactic_tokens(tokens):
    return is_comment_matching_pattern(token, regex)
  return False


def are_column_aligned(token_a, token_b):
  """
  Return true if both tokens are on the same column.
  """
  return token_a.begin.col == token_b.begin.col


def next_is_trailing_comment(config, tokens):
  """
  Return true if there is a trailing comment in the token stream
  """

  if not tokens:
    return False

  # If the next token is
  if next_is_explicit_trailing_comment(config, tokens):
    return True

  if is_valid_trailing_comment(tokens[0]):
    return True

  if len(tokens) < 2:
    return False

  if (tokens[0].type == lex.TokenType.WHITESPACE
      and is_valid_trailing_comment(tokens[1])):
    return True
  return False


def comment_is_tag(token):
  """
  Return true if the comment token has one of the tag-forms:

  # cmake-format: <tag>
  # cmf: <tag>
  #[[cmake-format:<tag>]]
  #[[cmf:<tag>]]
  """
  return get_tag(token) is not None


def pargs_are_full(npargs, nconsumed):
  if isinstance(npargs, int):
    return nconsumed >= npargs

  assert isinstance(npargs, STRING_TYPES), (
      "Unexpected npargs type {}".format(type(npargs)))

  if npargs == "?":
    return nconsumed >= 1

  if npargs in ("*", "+"):
    return False

  if npargs.endswith("+"):
    try:
      _ = int(npargs[:-1])
    except ValueError:
      raise ValueError("Unexepected npargs {}".format(npargs))
  return False


def npargs_is_exact(npargs):
  """
  Return true if npargs has an exact specification
  """
  return isinstance(npargs, int)


def get_min_npargs(npargs):
  """If the npargs specification string indicates a minimum number of arguments
     then return that value."""

  if npargs is None:
    return 0

  if isinstance(npargs, int):
    return npargs

  if npargs in ("*", "?"):
    return 0

  if npargs == "+":
    return 1

  if npargs.endswith("+"):
    try:
      return int(npargs[:-1])
    except ValueError:
      pass

  try:
    return int(npargs)
  except ValueError:
    pass

  raise ValueError(
      "Unexpected npargs {}({})".format(npargs, type(npargs).__name__))


LINE_TAG = re.compile(r"#\s*(cmake-format|cmf): ([^\n]*)")
BRACKET_TAG = re.compile(r"#\[(=*)\[(cmake-format|cmf):(.*)\]\1\]")


def get_tag(token):
  """
  If the token is a comment with one of the cmake-format tag forms, then
  extract the tag.
  """
  if token.type is lex.TokenType.COMMENT:
    match = LINE_TAG.match(token.spelling)
    if match:
      return match.group(2).strip().lower()
  elif token.type is lex.TokenType.BRACKET_COMMENT:
    match = BRACKET_TAG.match(token.spelling)
    if match:
      return match.group(3).strip().lower()

  return None


def get_normalized_kwarg(token):
  """
  Return uppercase token spelling if it is a word, otherwise return None
  """

  if (token.type == lex.TokenType.UNQUOTED_LITERAL
      and token.spelling.startswith('-')):
    return token.spelling.lower()

  if token.type != lex.TokenType.WORD:
    return None

  return token.spelling.upper()


def should_break(token, breakstack):
  """
  Return true if any function in breakstack evaluates to true on the current
  token. Otherwise return false.
  """
  for breakcheck in breakstack[::-1]:
    if breakcheck(token):
      return True
  return False


def only_comments_and_whitespace_remain(tokens, breakstack):
  skip_tokens = (lex.TokenType.WHITESPACE,
                 lex.TokenType.NEWLINE,
                 lex.TokenType.COMMENT,
                 lex.TokenType.BRACKET_COMMENT)

  for token in tokens:
    if token.type in skip_tokens:
      continue

    if should_break(token, breakstack):
      return True

    return False
  return True


def iter_syntactic_tokens(tokens):
  """
  Return a generator over the list of tokens yielding only those that are
  not whitespace
  """
  skip_tokens = (lex.TokenType.WHITESPACE,
                 lex.TokenType.NEWLINE)

  for token in tokens:
    if token.type in skip_tokens:
      continue
    yield token


def get_next_syntactic_token(tokens):
  """
  return the first non-whitespace token in the list
  """
  for token in iter_syntactic_tokens(tokens):
    return token
  return None


def iter_semantic_tokens(tokens):
  """
  Return a generator over the list of tokens yielding only those that
  have semantic meaning
  """

  for token in tokens:
    if is_semantic_token(token):
      yield token


def get_nth_semantic_token(tokens, nth):
  idx = 0
  for token in iter_semantic_tokens(tokens):
    if idx == nth:
      return token
    idx += 1
  return None


def get_first_semantic_token(tokens):
  """
  Return the first token with semantic meaning
  """
  return get_nth_semantic_token(tokens, 0)


def comment_belongs_up_tree(ctx, tokens, node, breakstack):
  """
  Return true if the comment token at tokens[0] belongs to some parent in the
  ancestry of `node` rather than to `node` itself. We determine this if the
  following is true:

  1. There is a semantic token remaining in the token stream
  2. The next semantic token would break the current scope of the node
  3. The column of the coment token is to the left of the column where the
     node starts

  e.g.

  ~~~
  statement_name(
    KEYWORD1 argument
             argument
             # comment 1
      # comment 2
    # comment 3
  )
  ~~~

  In this example:

  * comment 1 belongs to the positional group child of the keyword group
  * comment 2 belongs to the keyword group
  * comment 3 belongs to the statement
  """

  next_semantic = get_first_semantic_token(tokens)

  if next_semantic is None:
    # There is no next semantic token. We don't really expect this to happen
    return False

  if not should_break(next_semantic, breakstack):
    # The next token wouldn't break the current node
    return False

  if not ctx.argstack:
    # There is no parent node to assign the comment to, it must belong to the
    # current node
    return False

  return tokens[0].get_location().col < node.get_location().col


def parse_pspec(pargs, flags):
  """
  Parse a positional argument specification.
  """
  out = []

  # Default pargs is "*"
  if pargs is None:
    pargs = ZERO_OR_MORE

  # If we only have one scalar specification, return a legacy specification
  # TODO(josh): should we only do this if we also have flags?
  if isinstance(pargs, STRING_TYPES + (int,)):
    return [PositionalSpec(pargs, flags=flags, legacy=True)]

  if flags:
    raise UserError(
        "Illegal use of top-level 'flags' keyword with new-style positional"
        " argument declaration")

  # If we only have one dictionary specification, then put it in a dictionary
  # so that we can do the rest consistently
  if isinstance(pargs, dict):
    pargs = [pargs]

  for pargdecl in pargs:
    if isinstance(pargdecl, STRING_TYPES + (int,)):
      # A scalar declaration is interpreted as npargs
      out.append(PositionalSpec(pargdecl))
      continue

    if isinstance(pargdecl, dict):
        # A dictionary is interpreted as init kwargs
      if "npargs" not in pargdecl:
        pargdecl = dict(pargdecl)
        pargdecl["nargs"] = ZERO_OR_MORE
      out.append(PositionalSpec(**pargdecl))
      continue

    if isinstance(pargdecl, (list, tuple)):
        # A list or tuple is interpreted as (*args, [kwargs])
      args = list(pargdecl)
      kwargs = {}
      if isinstance(args[-1], dict):
        kwargs = args.pop(-1)
      out.append(PositionalSpec(*args, **kwargs))

  return out


class CommandSpec(object):
  """
  A command specification is composed of a sequence of positional argument
  specifications, and a dictionary mapping keyword arguments to (nested)
  command specifications. It also includes a command (or keyword) name.
  """

  def __init__(self, name, pargs=None, flags=None, kwargs=None, spelling=None,
               max_subgroups_hwrap=None, always_wrap=None):
    super(CommandSpec, self).__init__()
    scalar_types = (int,) + STRING_TYPES

    self.name = name
    self.spelling = spelling
    if spelling is None:
      self.spelling = name
    self.max_subgroups_hwrap = max_subgroups_hwrap
    self.always_wrap = always_wrap

    try:
      self.pargs = parse_pspec(pargs, flags)
    except (TypeError, UserError) as ex:
      message = (
          "Invalid user-supplied specification for positional arguments of "
          "{}:\n{}".format(name, ex))
      raise UserError(message)

    self.kwargs = {}

    if kwargs is not None:
      if isinstance(kwargs, dict):
        items = kwargs.items()
      elif isinstance(kwargs, (list, tuple)):
        items = list(kwargs)
      else:
        raise ValueError(
            "Invalid type {} for kwargs of {}: {}"
            .format(type(kwargs), name, kwargs))

      for keyword, spec in items:
        if isinstance(spec, scalar_types):
          self.kwargs[keyword] = CommandSpec(name=keyword, pargs=spec)
        elif isinstance(spec, CommandSpec):
          self.kwargs[keyword] = spec
        elif isinstance(spec, dict):
          self.kwargs[keyword] = CommandSpec(name=keyword, **spec)
        else:
          raise ValueError("Unexpected type '{}' for kwargs"
                           .format(type(kwargs)))

  def is_flag(self, key):
    return self.kwargs.get(key, None) == 0

  def is_kwarg(self, key):
    subspec = self.kwargs.get(key, None)
    if subspec is None:
      return False
    if isinstance(subspec, int):
      return subspec != 0
    if isinstance(subspec, STRING_TYPES + (CommandSpec,)):
      return True
    raise ValueError("Unexpected kwargspec for {}: {}"
                     .format(key, type(subspec)))

  def add(self, name, pargs=None, flags=None, kwargs=None, **extra):
    self.kwargs[name.lower()] = CommandSpec(
        name, pargs, flags, kwargs, **extra)


def apply_overrides(spectree, pathkeystr, value):
  """Apply command specification overrides to a previously constructed
     parse function tree."""

  if "." not in pathkeystr:
    raise UserError("Invalid override key {}".format(pathkeystr))
  pathstr, keystr = pathkeystr.rsplit(".", 1)

  pathparts = pathstr.split("/")

  while pathparts:
    if not isinstance(spectree, CommandSpec):
      raise InternalError("Invalid spectree entry of type {}"
                          .format(type(spectree)))

    pathkey = pathparts.pop(0)
    if pathkey.startswith("parg"):
      if pathparts:
        raise UserError(
            "Invalid override key {} contains path components after pspec"
            .format(pathkeystr))
      try:
        idx = int(pathkey[len("parg"):].strip("[]"))
      except ValueError:
        raise UserError(
            "Invalid override key {} contains non-int pspec"
            .format(pathkeystr))
      if idx > len(spectree.pargs):
        raise UserError(
            "Invalid override key {} is out of bounds"
            .format(pathkeystr))

      repl = {keystr: value}
      spectree.pargs[idx] = spectree.pargs[idx].replace(**repl)
      return

    if pathkey not in spectree.kwargs:
      raise UserError(
          "Invalid override key {} at {}".format(pathkeystr, pathkey))
    spectree = spectree.kwargs[pathkey]

  if not hasattr(spectree, keystr):
    raise UserError(
        "Invalid override key {}, keypart {} is not valid for type {}"
        .format(pathkeystr, keystr, type(spectree)))

  try:
    setattr(spectree, keystr, value)
  except AttributeError:
    raise UserError(
        "Invalid override key {}, can't set attribute {} for type {}"
        .format(pathkeystr, keystr, type(spectree)))



import io
from setuptools import setup

GITHUB_URL = "https://github.com/cheshirekow/cmakelang"

VERSION = None
with io.open("cmakelang/__init__.py", encoding="utf-8") as infile:
  for line in infile:
    line = line.strip()
    if line.startswith("__version__ ="):
      VERSION = line.split("=", 1)[1].strip().strip("'\"")

assert VERSION is not None


with io.open("cmakelang/doc/README.rst", encoding="utf-8") as infile:
  long_description = infile.read()

setup(
    name="cmakelang",
    packages=[
        "cmakelang",
        "cmakelang.format",
        "cmakelang.lex",
        "cmakelang.lint",
        "cmakelang.parse",
        "cmakelang.parse.funs",
    ],
    version=VERSION,
    description="Language tools for cmake (format, lint, etc)",
    long_description=long_description,
    author="Josh Bialkowski",
    author_email="josh.bialkowski@gmail.com",
    url=GITHUB_URL,
    download_url="{}/archive/{}.tar.gz".format(GITHUB_URL, VERSION),
    keywords=["cmake", "format"],
    license="GPLv3",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
    ],
    include_package_data=True,
    package_data={
        "cmakelang": [
            "templates/*"
        ]
    },
    entry_points={
        "console_scripts": [
            "cmake-annotate=cmakelang.annotate:main",
            "cmake-format=cmakelang.format.__main__:main",
            "cmake-lint=cmakelang.lint.__main__:main",
            "cmake-genparsers=cmakelang.genparsers:main",
            "ctest-to=cmakelang.ctest_to:main"
        ],
    },
    extras_require={
        "YAML": ["pyyaml>=5.3"],
        "html-gen": ["jinja2==2.10.3"]
    },
    install_requires=["six>=1.13.0"]
)

setup(
    name="cmake-annotate",
    packages=[],
    version=VERSION,
    description="Can format your listfiles so they don't look like crap",
    long_description=long_description,
    author="Josh Bialkowski",
    author_email="josh.bialkowski@gmail.com",
    url=GITHUB_URL,
    download_url="{}/archive/{}.tar.gz".format(GITHUB_URL, VERSION),
    keywords=["cmake", "annotate"],
    license="GPLv3",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
    ],
    include_package_data=True,
    install_requires=[
        "cmakelang>={}".format(VERSION),
        "pyyaml>=5.3"]
)

setup(
    name="cmake-format",
    packages=[],
    version=VERSION,
    description="Can format your listfiles so they don't look like crap",
    long_description=long_description,
    author="Josh Bialkowski",
    author_email="josh.bialkowski@gmail.com",
    url=GITHUB_URL,
    download_url="{}/archive/{}.tar.gz".format(GITHUB_URL, VERSION),
    keywords=["cmake", "format", "formatter"],
    license="GPLv3",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
    ],
    include_package_data=True,
    extras_require={
        "YAML": ["pyyaml>=5.3"],
    },
    install_requires=["cmakelang>={}".format(VERSION)]
)


setup(
    name="cmake-lint",
    packages=[],
    version=VERSION,
    description="Can check your listfiles for common problems",
    long_description=long_description,
    author="Josh Bialkowski",
    author_email="josh.bialkowski@gmail.com",
    url=GITHUB_URL,
    download_url="{}/archive/{}.tar.gz".format(GITHUB_URL, VERSION),
    keywords=["cmake", "lint", "linter"],
    license="GPLv3",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
    ],
    include_package_data=True,
    extras_require={
        "YAML": ["pyyaml>=5.3"],
    },
    install_requires=["cmakelang>={}".format(VERSION)]
)


setup(
    name="cmake-parse",
    packages=[],
    version=VERSION,
    description="Python library for parsing cmake listfiles. ",
    long_description=long_description,
    author="Josh Bialkowski",
    author_email="josh.bialkowski@gmail.com",
    url=GITHUB_URL,
    download_url="{}/archive/{}.tar.gz".format(GITHUB_URL, VERSION),
    keywords=["cmake", "lint", "linter"],
    license="GPLv3",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
    ],
    include_package_data=True,
    install_requires=["cmakelang>={}".format(VERSION)]
)



"""
Find all restructured text files and write them out to a manifest
"""

from __future__ import print_function

import argparse
import os

EXCLUDE_DIRS = [
    ".build",
    ".git"
]


def main():
  parser = argparse.ArgumentParser(description=__doc__)
  parser.add_argument("-m", "--manifest-path",
                      help="Path to the manifest file to create/update")
  parser.add_argument("-t", "--touch", action="store_true",
                      help="Touch the manifest if any rst files are newer")
  parser.add_argument("rootdir")
  args = parser.parse_args()

  rootdir = os.path.realpath(args.rootdir)

  latest_mtime = 0

  fileset = set()
  for parent, dirnames, filenames in os.walk(rootdir):
    dirnames[:] = sorted(dirnames)
    relpath_parent = os.path.relpath(parent, rootdir)
    if relpath_parent in EXCLUDE_DIRS:
      dirnames[:] = []
      continue

    for filename in filenames:
      if filename.endswith(".rst"):
        if relpath_parent == ".":
          relpath_file = filename
        else:
          relpath_file = os.path.join(relpath_parent, filename)
        file_mtime = os.path.getmtime(os.path.join(rootdir, relpath_file))
        latest_mtime = max(latest_mtime, file_mtime)
        fileset.add(relpath_file)

  manifest_mtime = 0
  manifest_path = os.path.realpath(args.manifest_path)
  manifest_set = set()
  if os.path.exists(manifest_path):
    manifest_mtime = os.path.getmtime(manifest_path)
    with open(manifest_path, "r") as infile:
      for line in infile:
        manifest_set.add(line.strip())

  if manifest_set != fileset:
    print("RST manifest has changed")
    with open(manifest_path, "w") as outfile:
      for relpath_file in sorted(fileset):
        outfile.write(relpath_file)
        outfile.write("\n")
  elif args.touch and latest_mtime > manifest_mtime:
    print("An RST file has changed")
    os.utime(manifest_path, None)


if __name__ == "__main__":
  main()



"""
Write out a pypirc file for use on travis. This allows us to use a different
password for each upload artifact, and between test.pypi.org and pypi.org
"""

import argparse
import io
import json

TEMPLATE = """\
[distutils]
index-servers =
  pypi
  testpypi

[pypi]
username=__token__
password={}

[testpypi]
repository=https://test.pypi.org/legacy/
username=__token__
password={}
"""


def main():
  argparser = argparse.ArgumentParser(description=__doc__)
  argparser.add_argument("--tokens-from")
  argparser.add_argument("--package")
  argparser.add_argument("-o", "--outfile")
  args = argparser.parse_args()

  with io.open(args.tokens_from, encoding="utf-8") as infile:
    data = json.load(infile)

  content = TEMPLATE.format(
      data["pypi.org"][args.package],
      data["test.pypi.org"][args.package])

  with io.open(args.outfile, "w", encoding="utf-8") as outfile:
    outfile.write(content)


if __name__ == "__main__":
  main()



"""
Test the include mechanism for loading configs
"""

from __future__ import print_function
from __future__ import unicode_literals

import io
import os
import shutil
import tempfile
import unittest

from cmakelang.format import __main__


class TestConfigInclude(unittest.TestCase):

  outdir = None

  def _list2reason(self, exc_list):
    if exc_list and exc_list[-1][0] is self:
      return exc_list[-1][1]
    return None

  def _test_passed(self):
    # Taken from https://stackoverflow.com/a/39606065/141023
    if hasattr(self, '_outcome'):  # Python 3.4+
      # these 2 methods have no side effects
      result = self.defaultTestResult()
      # pylint: disable=no-member
      self._feedErrorsToResult(result, self._outcome.errors)
    else:  # Python 3.2 - 3.3 or 3.0 - 3.1 and 2.7
      result = getattr(
          self, '_outcomeForDoCleanups',
          self._resultForDoCleanups)  # pylint: disable=no-member

    error = self._list2reason(result.errors)
    failure = self._list2reason(result.failures)
    return not error and not failure

    # # demo:   report short info immediately (not important)
    # if not ok:
    #   typ, text = ('ERROR', error) if error else ('FAIL', failure)
    #   msg = [x for x in text.split('\n')[1:] if not x.startswith(' ')][0]
    #   print("\n%s: %s\n     %s" % (typ, self.id(), msg))

  def setUp(self):
    self.outdir = tempfile.mkdtemp(
        prefix="cmake-format-{}-".format(self._testMethodName))

  def tearDown(self):
    if self._test_passed():
      shutil.rmtree(self.outdir)

  def test_single_include(self):
    outpath = os.path.join(self.outdir, "config-2.py")
    with io.open(outpath, "w") as outfile:
      outfile.write("""
var_a = "hello"
      """)

    outpath = os.path.join(self.outdir, "config-1.py")
    with io.open(outpath, "w") as outfile:
      outfile.write("""
include = ["config-2.py"]
var_b = "world"
      """)

    config_dict = __main__.get_configdict([outpath])
    self.assertIn("var_a", config_dict)
    self.assertIn("var_b", config_dict)

  def test_nested_include(self):
    outpath = os.path.join(self.outdir, "config-3.py")
    with io.open(outpath, "w") as outfile:
      outfile.write("""
var_a = "hello"
      """)

    outpath = os.path.join(self.outdir, "config-2.py")
    with io.open(outpath, "w") as outfile:
      outfile.write("""
include = ["config-3.py"]
var_b = "world"
      """)

    outpath = os.path.join(self.outdir, "config-1.py")
    with io.open(outpath, "w") as outfile:
      outfile.write("""
include = ["config-2.py"]
var_c = "foobar"
      """)

    config_dict = __main__.get_configdict([outpath])
    self.assertIn("var_a", config_dict)
    self.assertIn("var_b", config_dict)
    self.assertIn("var_c", config_dict)

  def test_relative_path(self):
    subdir_path = os.path.join(self.outdir, "subdir")
    os.makedirs(subdir_path)
    outpath = os.path.join(subdir_path, "config-2.py")
    with io.open(outpath, "w") as outfile:
      outfile.write("""
var_a = "hello"
      """)

    outpath = os.path.join(self.outdir, "config-1.py")
    with io.open(outpath, "w") as outfile:
      outfile.write("""
include = ["subdir/config-2.py"]
var_b = "world"
      """)

    config_dict = __main__.get_configdict([outpath])
    self.assertIn("var_a", config_dict)
    self.assertIn("var_b", config_dict)


if __name__ == "__main__":
  unittest.main()



# pylint: disable=W0613
import enum
import logging
import re

from cmakelang.common import InternalError
from cmakelang.format.formatter import get_comment_lines
from cmakelang.lex import TokenType, Token
from cmakelang.parse.argument_nodes import (
    ArgGroupNode, ConditionalGroupNode, PositionalGroupNode
)
from cmakelang.parse.body_nodes import BodyNode, FlowControlNode
from cmakelang.parse.simple_nodes import CommentNode
from cmakelang.parse.common import NodeType, TreeNode
from cmakelang.parse.statement_node import StatementNode
from cmakelang.parse.util import get_min_npargs
from cmakelang.parse import variables
from cmakelang.parse.funs.set import SetFnNode


logger = logging.getLogger(__name__)


def find_statements_in_subtree(subtree, funnames):
  """
  Return a generator that yields all statements in the `subtree` which match
  the provided set of `funnames`.
  """
  if isinstance(subtree, (list, tuple)):
    queue = subtree
  else:
    queue = [subtree]

  while queue:
    node = queue.pop(0)
    if isinstance(node, StatementNode):
      if node.get_funname() in funnames:
        yield node
    for child in node.children:
      if isinstance(child, TreeNode):
        queue.append(child)


def find_nodes_in_subtree(subtree, nodetypes):
  """Return a generator that yields all nodes in the `subtree` which are of the
     given type(s) `nodetypes`.
  """
  if isinstance(subtree, (list, tuple)):
    queue = subtree
  else:
    queue = [subtree]

  while queue:
    node = queue.pop(0)
    if isinstance(node, nodetypes):
      yield node

    if isinstance(node, TreeNode):
      for child in node.children:
        queue.append(child)


def loop_contains_argn(loop_stmt):
  """Return true if the loop statement contains ${ARGN} as an argument"""
  for token in loop_stmt.argtree.get_semantic_tokens():
    if token.type is TokenType.DEREF and token.spelling == "${ARGN}":
      return True
  return False


def statement_is_fundef(node):
  """Return true if a statement node is for a function or macro definition.
  """
  tokens = node.get_semantic_tokens()
  if not tokens:
    return False
  funname = tokens[0].spelling.lower()
  return funname in ("function", "macro")


def get_prefix_comment(prevchild):
  """
  Expect a sequence of COMMENT, WHITESPACE, <node>. If this sequence is true,
  return the comment node.
  """
  for idx in range(2):
    if not isinstance(prevchild[idx], TreeNode):
      return None
  if not prevchild[0].node_type is NodeType.WHITESPACE:
    return None

  newline_count = 0
  for token in prevchild[0].get_tokens():
    newline_count += token.spelling.count("\n")
  if newline_count > 1:
    return None

  if not prevchild[1].node_type is NodeType.COMMENT:
    return None
  return prevchild[1]


def get_list_outvar(node):
  """Given a statement parse node for a `list()` command, inspect the
     subcommand and extract the token corresponding to the output
     variable name."""
  semtoks = node.argtree.parg_groups[0].get_tokens(kind="semantic")
  listcmd = semtoks[0].spelling
  if listcmd.upper() in ("LENGTH", "GET", "JOIN", "SUBLIST"):
    return semtoks[-1]
  return None


class Scope(enum.IntEnum):
  CACHE = 0  # global scope (public)
  INTERNAL = 1  # global scope (private)
  PARENT = 2  # indeterminate scope, but it can't be CACHE
  LOCAL = 3  # function local scope
  DIRECTORY = 4  # directory scope
  LOOP = 5  # loop variable name
  ARGUMENT = 6  # argument variable name


def get_scope_of_assignment(set_node):
  """Return the scope of the assignment."""
  args = set_node.argtree
  if isinstance(args, SetFnNode):
    if args.cache:
      if args.cache.type == "INTERNAL":
        return Scope.INTERNAL
      return Scope.CACHE

    if args.parent_scope:
      return Scope.PARENT

  # assume block (directory) scope
  prev = set_node
  parent = args.parent
  while parent:
    if isinstance(parent, FlowControlNode):
      block = parent.get_block_with(prev)
      if block.open_stmt.get_funname() == "function":
        # found ourselves in the body of a function, which means that we are
        # local scope
        return Scope.LOCAL
      if block.open_stmt.get_funname() == "macro":
        # found ourselves in the body of a macro, which means that we are
        # in parent scope
        return Scope.PARENT
    prev = parent
    parent = parent.parent
  return Scope.DIRECTORY


def make_varref_callback(repl):
  """Return a function for regex substitution to replace a variable
     substitution regex with a stub string."""

  def varref_callback(matchobj):
    if matchobj.group(1):
      return matchobj.group(1) + repl
    return repl

  return varref_callback


def mock_varrefs(tokenstr, repl=None):
  """Recursively replace variable references with a dummy string until all
     variable references are resolved.

  :see: https://cmake.org/cmake/help/v3.12/policy/CMP0053.html#policy:CMP0053
  """

  if repl is None:
    repl = "foo"

  varref = re.compile(
      r"(?<!\\)(\\\\)*"
      r"\$\{(?:(?:[A-Za-z0-9_./+-])|(?:\\[^A-Za-z0-9_./+-]))+\}")

  # NOTE(josh): in python2 we are not allowed to use \1 when the match is
  # empty, so we have to use a callback function to implement the replacement.
  while varref.search(tokenstr):
    tokenstr = varref.sub(make_varref_callback(repl), tokenstr)
  return tokenstr


class LintChecker(object):
  def __init__(self, cfg, local_ctx):
    self.cfg = cfg
    self.local_ctx = local_ctx
    self._node_stack = []
    self._indent_token = None

  @property
  def context(self):
    """
    Convenience shim to avoid a big diff in the code. Before this was a class
    each checker take in a `(cfg, local_ctx)` as input arguments. Instead of
    replacing each usage of these arguments, we just do
    `(cfg, local_ctx) = self.context` at the start of each method and keep the
    rest of the code the same.
    """
    return (self.cfg, self.local_ctx)

  def check_basics(self, infile_content):
    """Perform  basic checks before even lexing the file
    """
    (cfg, local_ctx) = self.context
    lines = infile_content.split("\n")
    indent_regex = re.compile(r"([ \t]*)(.*)")

    for lineno, line in enumerate(lines):
      # NOTE(josh): The tokenizer starts lineno at "1", so we must do the same
      # here in order to be consistent.
      lineno += 1
      if len(line) > cfg.format.line_width:
        local_ctx.record_lint(
            "C0301", len(line), cfg.format.line_width, location=(lineno,))

      if line.endswith("\r"):
        if cfg.format.line_ending == "unix":
          local_ctx.record_lint(
              "C0327", "windows", location=(lineno,))
        line = line[:-1]
      else:
        if cfg.format.line_ending == "windows":
          local_ctx.record_lint(
              "C0327", "unix", location=(lineno,))

      if len(line.rstrip()) != len(line):
        local_ctx.record_lint("C0303", location=(lineno,))

      match = indent_regex.match(line)
      indentation = match.group(1)
      if not cfg.format.use_tabchars:
        if "\t" in indentation:
          colno = indentation.find("\t")
          local_ctx.record_lint(
              "C0306", "tab", "space", location=(lineno, colno))
      elif cfg.format.fractional_tab_policy == "round-up":
        if " " in indentation:
          colno = indentation.find(" ")
          local_ctx.record_lint(
              "C0306", "space", "tab", location=(lineno, colno))
      else:
        subindent = indentation.lstrip("\t")
        if "\t" in subindent:
          colno = len(indentation) - len(subindent)
          local_ctx.record_lint(
              "C0306", "space", "tab", location=(lineno, colno))

    # check that the file ends with newline
    if not infile_content.endswith("\n"):
      local_ctx.record_lint("C0304", location=(len(lines),))

  def check_tokens(self, tokens):
    """Look for anything that looks like an incomplete variable substitution."""
    (_, local_ctx) = self.context

    missing_suffix = re.compile(
        r"(?<!\\)(?:\\\\)*"
        r"(\$\{)((?:(?:[A-Za-z0-9_./+-])|(?:\\[^A-Za-z0-9_./+-]))+)")

    missing_prefix = re.compile(
        r"(?<!\\)(?:\\\\)*"
        r"(\$|\{)((?:(?:[A-Za-z0-9_./+-])|(?:\\[^A-Za-z0-9_./+-]))+)")

    match_types = (
        TokenType.QUOTED_LITERAL,
        TokenType.UNQUOTED_LITERAL,
        TokenType.BRACKET_ARGUMENT
    )

    for token in tokens:
      if token.type not in match_types:
        continue

      resolved = mock_varrefs(token.spelling)

      match = missing_prefix.search(resolved)
      if match and variables.CASE_SENSITIVE_REGEX.match(match.group(2)):
        catmatch = "".join(match.group(1, 2))
        if catmatch == "$ENV":
          # This is an environment variable reference, so we don't try to
          # match it
          continue

        local_ctx.record_lint(
            "W0106", "open", catmatch, location=token.get_location())
        continue

      match = missing_suffix.search(resolved)
      if match and variables.CASE_SENSITIVE_REGEX.match(match.group(2)):
        local_ctx.record_lint(
            "W0106", "closing", catmatch, location=token.get_location())
        continue

  def check_for_custom_parse_logic(self, stmt_node):
    """Ensure that a function or macro definition doesn't contain custom parser
      logic. The check is heuristic, but what we look for is a loop over ARGN
      where the body of the loop contains multiple conditional checks against
      the string value of the arguments
    """
    (cfg, local_ctx) = self.context
    block = stmt_node.parent.get_block_with(stmt_node)
    for _ in find_statements_in_subtree(
        block.body, ("cmake_parse_arguments",)):
      # function/macro definition uses the std parser, so the check is complete
      return

    for loop_stmt in find_statements_in_subtree(
        block.body, ("foreach", "while")):
      if loop_contains_argn(loop_stmt):
        conditional_count = 0
        loopvar = loop_stmt.argtree.get_semantic_tokens()[0]
        loop_body = loop_stmt.parent.get_block_with(loop_stmt).body
        for conditional in find_nodes_in_subtree(
            loop_body, ConditionalGroupNode):
          tokens = conditional.get_semantic_tokens()
          if not tokens:
            continue
          if tokens[0].spelling == loopvar.spelling:
            if tokens[1].spelling in ("STREQUAL", "MATCHES"):
              conditional_count += 1
        if conditional_count > cfg.lint.max_conditionals_custom_parser:
          local_ctx.record_lint("C0201", location=loop_stmt.get_location())
          return

  def check_argument_names(self, defn_node):
    """Check that the argument names in a function or macro definition match
      the required pattern."""
    (cfg, local_ctx) = self.context
    tokens = defn_node.argtree.get_semantic_tokens()[1:]
    seen_names = set()
    uncase_names = set()
    for token in tokens:  # named arguments
      if token.type is TokenType.RIGHT_PAREN:
        break
      if token.type is not TokenType.WORD:
        local_ctx.record_lint(
            "E0109", token.spelling, location=token.get_location())
      else:
        if token.spelling in seen_names:
          local_ctx.record_lint(
              "E0108", token.spelling, location=token.get_location())
        elif token.spelling.lower() in uncase_names:
          local_ctx.record_lint(
              "C0202", token.spelling, location=token.get_location())
        elif not re.match(cfg.lint.argument_var_pattern, token.spelling):
          local_ctx.record_lint(
              "C0103", "argument", token.spelling,
              cfg.lint.argument_var_pattern, location=token.get_location())
        seen_names.add(token.spelling)
        uncase_names.add(token.spelling.lower())

    if len(tokens) > cfg.lint.max_arguments:
      local_ctx.record_lint("R0913", len(tokens), cfg.lint.max_arguments,
                            location=defn_node.get_location())

  def check_name_against_pattern(self, defn_node, pattern):
    """Check that a function or macro name matches the required pattern."""
    (_, local_ctx) = self.context
    tokens = defn_node.get_semantic_tokens()
    funname = tokens.pop(0).spelling  # statement name "function" or "macro"
    tokens.pop(0)  # lparen

    token = tokens.pop(0)  # function/macro name
    if not re.match(pattern, token.spelling):
      local_ctx.record_lint(
          "C0103", funname.lower(), token.spelling, pattern,
          location=token.get_location())

  def check_defn(self, defn_node, name_pattern):
    """Perform checks on a function or macro"""
    (cfg, local_ctx) = self.context
    self.check_for_custom_parse_logic(defn_node)
    self.check_name_against_pattern(defn_node, name_pattern)
    self.check_argument_names(defn_node)

    # TODO(josh): I guess logically the following checks should really be
    # part of check_body... in the case that the body within a function.
    # I'm not sure what the best way to organize that flow. In any case
    # we'll report the following errors at the first line following the
    # opening statement, which is where the body starts
    body_line = defn_node.get_tokens()[-1].get_location().line + 1

    block = defn_node.parent.get_block_with(defn_node)
    return_count = sum(
        1 for _ in find_statements_in_subtree(block.body, ("return",)))
    if return_count > cfg.lint.max_returns:
      local_ctx.record_lint("R0911", return_count, cfg.lint.max_returns,
                            location=(body_line,))

    branch_count = sum(
        1 for _ in find_statements_in_subtree(
            block.body, ("if", "elseif", "else")))
    if branch_count > cfg.lint.max_branches:
      local_ctx.record_lint("R0912", branch_count, cfg.lint.max_branches,
                            location=(body_line,))

    stmt_count = sum(
        1 for _ in find_nodes_in_subtree(block.body, StatementNode))
    if stmt_count > cfg.lint.max_statements:
      local_ctx.record_lint("R0915", stmt_count, cfg.lint.max_statements,
                            location=(body_line,))

  def check_fundef(self, node):
    """Perform checks on a function definition"""
    (cfg, _) = self.context
    self.check_defn(node, cfg.lint.function_pattern)

  def check_macrodef(self, node):
    """Perform checks on a macro definition"""
    (cfg, _) = self.context
    self.check_defn(node, cfg.lint.macro_pattern)

  def check_foreach(self, node):
    """Make sure that the loop variable matches the function argument
       pattern."""
    (cfg, local_ctx) = self.context

    tokens = node.get_semantic_tokens()
    tokens.pop(0)  # statement name "foreach"
    tokens.pop(0)  # lparen
    token = tokens.pop(0)  # loopvaraible
    if not re.match(cfg.lint.argument_var_pattern, token.spelling):
      local_ctx.record_lint(
          "C0103", "argument", token.spelling, cfg.lint.argument_var_pattern,
          location=token.get_location())

  def check_flow_control(self, node):
    """Perform checks on a flowcontrol node."""
    stmt = node.children[0]
    funname = stmt.children[0].children[0].spelling.lower()
    if funname == "function":
      self.check_fundef(stmt)
    elif funname == "macro":
      self.check_macrodef(stmt)
    elif funname == "foreach":
      self.check_foreach(stmt)

  def parse_pragmas_from_token(self, content, row, col, suppressions):
    """Parse any cmake-lint directives (pragmas) from line-comment
      tokens at the current scope."""

    (_, local_ctx) = self.context
    items = content.split()
    for item in items:
      if "=" not in item:
        local_ctx.record_lint("E0011", item, location=(row, col))
        col += len(item) + 1
        continue

      key, value = item.split("=", 1)
      if key == "disable":
        idlist = value.split(",")
        for idstr in idlist:
          if local_ctx.is_idstr(idstr):
            suppressions.append(idstr)
          else:
            local_ctx.record_lint(
                "E0012", idstr, location=(row, col))
      else:
        local_ctx.record_lint(
            "E0011", key, location=(row, col))

  def parse_pragmas_from_comment(self, node):
    """Parse any cmake-lint directives (pragmas) from line comment tokens within
      the comment node."""
    suppressions = []
    for child in node.children:
      if not isinstance(child, Token):
        continue
      if child.type is not TokenType.COMMENT:
        continue
      token = child
      pragma_prefix = "# cmake-lint: "
      if not token.spelling.startswith(pragma_prefix):
        continue

      row, col, _ = token.get_location()
      content = token.spelling[len(pragma_prefix):]
      col += len(pragma_prefix)
      self.parse_pragmas_from_token(content, row, col, suppressions)

    return suppressions

  def check_body(self, node):
    """Perform checks on a body node."""
    (cfg, local_ctx) = self.context
    suppressions = []
    prevchild = [None, None]
    for idx, child in enumerate(node.children):
      if not isinstance(child, TreeNode):
        # Should not be the case. Should we assert here?
        continue

      if child.node_type is NodeType.COMMENT:
        requested_suppressions = self.parse_pragmas_from_comment(child)
        if requested_suppressions:
          lineno = child.get_tokens()[0].get_location().line
          new_suppressions = local_ctx.suppress(lineno, requested_suppressions)
          suppressions.extend(new_suppressions)

      # Check for docstrings
      # TODO(josh): move into flow-control or fundef/macrodef checkers? Would
      # require an API to get siblings
      if child.node_type is NodeType.FLOW_CONTROL:
        stmt = child.children[0]
        if statement_is_fundef(stmt):
          prefix_comment = get_prefix_comment(prevchild)
          if not prefix_comment:
            local_ctx.record_lint(
                "C0111", location=child.get_semantic_tokens()[0].get_location())
          elif not "".join(get_comment_lines(cfg, prefix_comment)).strip():
            local_ctx.record_lint(
                "C0112", location=child.get_semantic_tokens()[0].get_location())

      # Check spacing between statements at block level
      # TODO(josh): move into flow-control and statement checkers? Would
      # require an API to get siblings
      if child.node_type in (NodeType.FLOW_CONTROL, NodeType.STATEMENT):
        if prevchild[0] is None:
          pass
        elif prevchild[0].node_type is not NodeType.WHITESPACE:
          local_ctx.record_lint("C0321", location=child.get_location())
        elif prevchild[0].count_newlines() < 1:
          local_ctx.record_lint("C0321", location=child.get_location())
        elif prevchild[0].count_newlines() < cfg.lint.min_statement_spacing:
          local_ctx.record_lint(
              "C0305", "not enough", location=child.get_location())
        elif prevchild[0].count_newlines() > cfg.lint.max_statement_spacing:
          local_ctx.record_lint(
              "C0305", "too many", location=child.get_location())

      if (isinstance(child, StatementNode) and
          child.get_funname() in ("break", "continue", "return")):
        for _ in find_nodes_in_subtree(node.children[idx + 1:], StatementNode):
          local_ctx.record_lint("W0101", location=child.get_location())
          break

      prevchild[1] = prevchild[0]
      prevchild[0] = child

    lineno = node.get_tokens()[-1].get_location().line
    if suppressions:
      local_ctx.unsuppress(lineno, suppressions)

  def check_arggroup(self, node):
    (_, local_ctx) = self.context
    kwargs_seen = set()
    for child in node.children:
      if isinstance(child, TreeNode) and child.node_type is NodeType.KWARGGROUP:
        kwarg_token = child.get_semantic_tokens()[0]
        kwarg = kwarg_token.spelling.upper()
        if kwarg in ("AND", "OR", "COMMAND", "PATTERN", "REGEX"):
          continue
        if kwarg in kwargs_seen:
          local_ctx.record_lint(
              "E1122", kwarg, location=kwarg_token.get_location())
        kwargs_seen.add(kwarg)

  def check_positional_group(self, node):
    """Perform checks on a positional group node."""
    (_, local_ctx) = self.context
    if node.spec is None:
      raise InternalError("Missing node.spec for {}".format(node))
    min_npargs = get_min_npargs(node.spec.npargs)
    semantic_tokens = node.get_semantic_tokens()
    if len(semantic_tokens) < min_npargs:
      location = ()
      if semantic_tokens:
        location = semantic_tokens[0].get_location()
      local_ctx.record_lint("E1120", location=location)

  def check_is_in_loop(self, node):
    """Ensure that a break() or continue() statement has a foreach() or
      while() node in it's ancestry."""
    (_, local_ctx) = self.context
    prevparent = node
    parent = node.parent

    while parent:
      if not isinstance(parent, FlowControlNode):
        prevparent = parent
        parent = parent.parent
        continue

      block = parent.get_block_with(prevparent)
      if block is None:
        prevparent = parent
        parent = parent.parent
        continue

      if block.open_stmt.get_funname() in ("foreach", "while"):
        return True

      prevparent = parent
      parent = parent.parent
    local_ctx.record_lint(
        "E0103", node.get_funname(), location=node.get_location())

  def check_assignment(self, node):
    """Checks on a variable assignment."""
    (cfg, local_ctx) = self.context
    scope = get_scope_of_assignment(node)

    if node.get_funname() == "set":
      varname = node.argtree.varname
    elif node.get_funname() == "list":
      varname = get_list_outvar(node)
      if varname is None:
        return
    else:
      logger.warning("Unexpected node %s (%s)", node, node.get_funname())
      return

    resolved = mock_varrefs(varname.spelling, "")

    # The variable name is hidden behind a variable reference so we can't
    # check it until we have a model for variables in scope.
    if not resolved:
      return

    if scope is Scope.CACHE:
      # variable is global scope and public
      pattern = cfg.lint.global_var_pattern
      if not re.match(pattern, resolved):
        local_ctx.record_lint(
            "C0103", "CACHE variable", varname.spelling, pattern,
            location=varname.get_location())
    elif scope is Scope.INTERNAL:
      # variable is global scope but private
      pattern = cfg.lint.internal_var_pattern
      if not re.match(pattern, resolved):
        local_ctx.record_lint(
            "C0103", "INTERNAL variable", varname.spelling,
            cfg.lint.internal_var_pattern,
            location=varname.get_location())
    elif scope is Scope.PARENT:
      # indeterminate scope, but it's not global
      pattern = "|".join([
          cfg.lint.public_var_pattern,
          cfg.lint.private_var_pattern,
          cfg.lint.local_var_pattern,
      ])

      if not re.match(pattern, resolved):
        local_ctx.record_lint(
            "C0103", "PARENT_SCOPE variable", varname.spelling, pattern,
            location=varname.get_location())
    elif scope is Scope.LOCAL:
      pattern = cfg.lint.local_var_pattern
      if not re.match(pattern, resolved):
        local_ctx.record_lint(
            "C0103", "local variable", varname.spelling,
            pattern, location=varname.get_location())
    elif scope is Scope.DIRECTORY:
      # We cannot tell by assignment whether it's meant to be public or private,
      # but it must match one of these patterns
      pattern = "|".join([
          cfg.lint.public_var_pattern,
          cfg.lint.private_var_pattern,
      ])
      if not re.match(pattern, resolved):
        local_ctx.record_lint(
            "C0103", "directory variable", varname.spelling, pattern,
            location=varname.get_location())

  def check_statement(self, node):
    """Perform checks on a statement."""
    if node.get_funname() in ("break", "continue"):
      self.check_is_in_loop(node)
    elif node.get_funname() in ["set", "list"]:
      self.check_assignment(node)

  def check_varname(self, varname, token, contextstr):
    """
    Record lint if the varname is a case-insensitive match to any builtin
    variable names, meaning that the author likely made a spelling mistake.
    """
    (_, local_ctx) = self.context

    imatch = variables.CASE_INSENSITIVE_REGEX.match(varname)
    if not imatch:
      # variable name isn't a match for any builtins
      return

    for groupstr in imatch.groups():
      if groupstr is not None:
        # variable name matches a dynamic pattern and we don't have the context
        # yet to actually compare against these
        return

    if not variables.CASE_SENSITIVE_REGEX.match(varname):
      # variable name is a match for a builtin except for case
      local_ctx.record_lint(
          "W0105", contextstr, varname,
          location=token.get_location())

  def check_variable_assignments(self, tree):
    """
    Check if any variable assignments are a case-insensitive match to any
    builtin variable names. This is probably a spelling error.
    """
    for stmt in find_statements_in_subtree(tree, ["set", "list"]):
      if stmt.get_funname() == "set":
        token = stmt.argtree.varname
      elif stmt.get_funname() == "list":
        token = stmt.argtree.parg_groups[0].get_tokens(kind="semantic")[1]
      else:
        continue
      self.check_varname(token.spelling, token, "Assignment to")

  def check_variable_references(self, tree):
    """
    Check if any variable references are a case-insensitive match to any
    builtin variable names. This is probably a spelling error.
    """
    # TODO(josh): replace with a stateful parser that builds up
    # global/directory/local namespaces and can check for usage before
    # assignment, shadowing, etc
    for token in tree.get_tokens(kind="semantic"):
      if token.type not in (
          TokenType.QUOTED_LITERAL, TokenType.DEREF):
        continue
      for varname in re.findall(r"\$\{([\w_]+)\}", token.spelling):
        self.check_varname(varname, token, "Reference to")

  def check_comment(self, node):
    if not self.am_in_statement():
      return

    requested_suppressions = self.parse_pragmas_from_comment(node)
    if not requested_suppressions:
      return

    (_cfg, local_ctx) = self.context
    lineno = node.get_tokens()[0].get_location().line
    new_suppressions = local_ctx.suppress(lineno, requested_suppressions)
    local_ctx.unsuppress(lineno + 1, new_suppressions)

  def am_in_statement(self):
    for node in self._node_stack:
      if isinstance(node, StatementNode):
        return True
    return False

  def am_in_arggroup(self):
    for node in self._node_stack:
      if isinstance(node, ArgGroupNode):
        return True
    return False

  def get_scope_depth(self):
    depth = 0
    for node in self._node_stack:
      if isinstance(node, BodyNode):
        depth += 1
    return depth

  def check_token(self, token):
    (cfg, local_ctx) = self.context

    if token.type is TokenType.WHITESPACE and token.get_location().col == 0:
      # A whitespace token at column zero is an intentation. It's spelling is
      # composed of whitespace characters (space, tab) composing the indent.
      self._indent_token = token
      return

    if self._indent_token:
      # This is the first non-whitespace token on this line
      desired_indentation = (self.get_scope_depth() - 1) * cfg.format.tab_size

      # Instead of using the column, we could get the previous token in the
      # token stream
      actual_indentation = (
          len(self._indent_token.spelling) +
          self._indent_token.spelling.count("\t") * (cfg.format.tab_size - 1))

      if self.am_in_arggroup():
        if actual_indentation < desired_indentation:
          local_ctx.record_lint(
              "C0307", self._indent_token.spelling, token.spelling,
              " " * desired_indentation, ">", location=token.get_location())
      else:
        if actual_indentation != desired_indentation:
          msg = "->".join([str(x) for x in self._node_stack])
          local_ctx.record_lint(
              "C0307", self._indent_token.spelling, token.spelling,
              " " * desired_indentation, msg, location=token.get_location())

      self._indent_token = None

  def check_tree(self, node):
    if isinstance(node, Token):
      self.check_token(node)
      return

    if not isinstance(node, TreeNode):
      return

    self._node_stack.append(node)
    if node.node_type is NodeType.BODY:
      self.check_body(node)
    elif node.node_type is NodeType.FLOW_CONTROL:
      self.check_flow_control(node)
    elif isinstance(node, ArgGroupNode):
      self.check_arggroup(node)
    elif isinstance(node, StatementNode):
      self.check_statement(node)
    elif isinstance(node, PositionalGroupNode):
      self.check_positional_group(node)
    elif isinstance(node, CommentNode):
      self.check_comment(node)

    for child in node.children:
      self.check_tree(child)
    self._node_stack.pop(-1)

  def check_parse_tree(self, node):
    self.check_variable_assignments(node)
    self.check_variable_references(node)
    self.check_tree(node)



"""
Database of lint

Categories mirror pylint:

* (C) convention, for programming standard violation
* (R) refactor, for bad code smell
* (W) warning, for cmake specific problems
* (E) error, for most probably bugs in the code
* (F) fatal, if an error occurred which prevented cmake-lint from doing further
      processing.
"""
from __future__ import unicode_literals


class Lint(object):
  """Basically a named-tuple to store a lint id, message format, and metadata
  """

  def __init__(self, idstr, msgfmt, description=None, explain=None):
    self.idstr = idstr
    self.msgfmt = msgfmt
    self.description = description
    self.explain = explain


def get_database():
  return {idstr: Lint(idstr, msgfmt, **kwargs)
          for idstr, msgfmt, kwargs in LINT_DB}


LINT_DB = [
# pylint:disable=bad-continuation
# noqa: E122
(
"C0102", "Black listed name \"{:s}\"", {
"description": """\
Used when the name is listed in the "bad-names" black list.

This message belongs to the basic checker.
""",
"explain": """\
cmake-lint can be customized to help enforce coding guidelines that discourage
or forbid use of certain names for variables, functions, etc.. These names are
specified with the bad-names option. This message is raised whenever a name is
in the list of names defined with the bad-names option.
"""
}), (
"C0103", "Invalid {:s} name \"{:s}\" doesn't match `{:s}`", {
"description": """\
Used when a name doesn't doesn't fit the naming convention associated to its
type (function, macro, variable, ...).

This message belongs to the basic checker.
""",
"explain": """\
The naming convention is defined with a regular expression, and the naming
convention is satisfied if the name matches the regular expression.
"""
}), (
"C0111", "Missing docstring on function or macro declaration", {
"description": """\
Used when a function or macro is defined without a documentation comment
immediately preceeding it.

This message belongs to the basic checker.
""",
"explain": """\
So, you've written a some fancy function that makes it "easier" to declare build
steps. Congratulations. You probably shouldn't have, but thats OK. Now that you
did, how should people use it? What arguments does it take? What are the
semantics of those arguements? You should include documentation in a comment
block prior to the function declaration with this information.
"""
}), (
"C0112", "Empty docstring on function or macro declaration", {
"description": """\
Used when a function or macro is preceeded by an empty comment string, rather
that one with useful documentation.

This message belongs to the basic checker.
""",
"explain": """\
Ok so you saw C0111 and figured you'd be clever right? Sorry, no dice. Please
include some useful documentation so that code readers know what your
function/macro does and how to use it.
"""
}), (
"C0113", "Missing {:s} in statement which allows it", {
}), (
"C0114", "Form descriminator hidden behind variable dereference", {
"description": """\
Used when a keyword used to descriminate betwen different forms of a command is
hidden behind a variable dereference.

This message is implemented by individual command checkers.
""",
"explain": """\
Some cmake commands have very different behavior depending on the presence of
a particular keyword (see e.g. the `file` command). And because cmake is a
macro languge that keyword can actually be held inside a variable. Thus the
keyword might not actually be visible to cmake-lint (or humans). In general
there is no reason to do this and it really hurts readability since different
descriminator keywords yield essentially different commands.
"""
}), (
"C0201", "Consider replacing custom parser logic with cmake_parse_arguments", {
"description": """\
Used when custom parse logic is detected.
"""
}), (
"C0202", "Argument name {:s} differs from existing argument only in case", {
}), (
"C0301", "Line too long ({:d}/{:d})", {
"description": """\
Used when a line is longer than the limit specified in the line-length
option.
""",
"explain": """\
It is a good idea to keep each line within a maximum length to keep it from
wrapping past the edge of an editing window. This improves readability and
tempers other developers' irritability!

The default value of the line-length option is 80, the customary width of a
terminal window.

Note that the line length and the limit are counted in characters, not in Bytes
needed to represent these characters.
"""
}), (
"C0303", "Trailing whitespace", {
"description": """\
Used when a line has one or more whitespace characters directly before the line
end character(s).

This message belongs to the basic checker.
""",
"explain": """\
Such trailing whitespace is visually indistinguishable and some editors will
trim them.
"""
}), (
"C0304", "Final newline missing", {
"description": """\
Used when a listfile has no line end character(s) on its last line.

This message belongs to the basic checker.
""",
"explain": """\
While cmake itself does not require line end character(s) on the last line,
is simply good practice to have it.
"""
}), (
"C0305", "{:s} newlines between statements", {
}), (
"C0306", "Tab-policy violation. Found {:s} but should be {:s}", {
}), (
"C0307", "Bad indentation:\n{:s}{:s}\n{:s}^----{}\n", {
}), (
"C0321", "Multiple statements on a single line", {
}), (
"C0327", "Wrong line ending ({:s})", {
"description": """\
Used when a line ends with the wrong line ending character. e.g. A line ends
with "\\r\\n" when configured for "\\n".

This message belongs to the basic checker.
""",
"explain": """\
While cmake itself does not enforce a particular line ending, it is good
practice for a project to be consist with their line endings.
"""
}), (
"E0011", "Unrecognized file option {:s}", {
"description": """\
Used when an unrecognized pragma is encountered.
""",
"explain": """\
cmake-lint allows for some inline comments to supress warnings (among other
things). This lint is emitted if a bad option key is provided in such a pragma
"""
}), (
"E0012", "Bad option value {:s}", {
"description": """\
Used when a cmake-lint pragma is encountered which attempts to alter some option
in an invalid way.

This message belongs to the basic checker.
""",
"explain": """\
cmake-lint allows for some inline comments to supress warnings (among other
things). This lint is emitted if a bad option is provided to one of these
pragmas.
"""
}), (
"E0103", "{:s} outside of loop", {
"description": """\
Used when a break() or continue() statement is used outside a loop.

This message belongs to the basic checker.
""",
}), (
"E0108", "Duplicate argument name {:s} in function/macro definition", {
}), (
"E0109", "Invalid argument name {:s} in function/macro definition", {
}), (
"E1120", "Missing required positional argument", {
"description": """\
Used when a positional argument group expecting an exact number of arguments
is closed (by a parenthesis) before that number of arguments is found.

This message belongs to the basic checker.
""",
}), (
"E1121", "Too many positional arguments", {
"description": """\
Used when a positional argument is found when no argument group is expected.

This message is implemented by individual command checkers
""",
}), (
"E1122", "Duplicate keyword argument {:s}", {
"description": """
Used when a keyword shows up more than once within an argument group. In
general, only COMMAND is allowed more than once.
"""
}), (
"E1125", "Missing required keyword argument {:s}", {
}), (
"E1126", "Invalid form descriminator", {
"description": """
Used when a keyword used to descriminate between different command forms is
ommitted.
"""
}), (
"R0911", "Too many return statements {:d}/{:d}", {
  }), (
"R0912", "Too many branches {:d}/{:d}", {
  }), (
"R0913", "Too many named arguments {:d}/{:d}", {
  }), (
"R0914", "Too many local variables {:d}/{:d}", {
  }), (
"R0915", "Too many statements {:d}/{:d}", {
}), (
"W0101", "Unreachable code", {
}), (
"W0104", "Use of deprecated command {:s}", {
}), (
"W0105",
"{:s} variable '{:s}' which matches a built-in except for case", {
"description": """
This warning means that you are using a variable such as,
for example, `cmake_cxx_standard` which matches a builtin variable
(`CMAKE_CXX_STANDARD`) except for the case. If this was intentional, then it's
bad practice as it causes confusion (there are two variables in the namespace
with identical name except for case), though it was probably not intentional
and you probably aren't assigning to the correct variable.

This warning may be emitted for assignment (e.g. `set()` or `list()`) as
well as for variable expansion in an argument (e.g. `"${CMAKE_Cxx_STANDARD}"`).
"""
}), (
"W0106",
"String looks like a variable reference missing an {:s} tag '{:s}'", {
})
]
